/********************************************************************
        example2.c - example of use of XCDE API - July 2002
	        Andrea Mastroianni - Paolo Ferragina 
      Dipartimento di Informatica - Universita' di Pisa (Italy)
	               ferragin@di.unipi.it
                       mastroia@di.unipi.it

   for permissions of distribution and use see the COPYRIGHT file
   contained in the XCDE package
********************************************************************/

// header file that must be always included in a program
// that uses the XCDE library.
#include"xcde_lib.h"

#define MACRO_HEADER { \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fprintf(stdout, "\n\t\tXCDE Library example of use") ;           \
fprintf(stdout, "\n\tby Paolo Ferragina - Andrea Mastroianni") ; \
fprintf(stdout, "\n\t\b\bferragin@di.unipi.it - mastroia@di.unipi.it") ; \
fprintf(stdout, "\n\t        Dipartimento di Informatica") ;     \
fprintf(stdout, "\n\t           Universita' di Pisa") ;          \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fflush(stdout) ; \
} 

#define MACRO_HELP { \
fprintf(stdout, "\n\nSYNTAX: example2 [FILENAME] [TAG] [WORD] [WORD]") ; \
fprintf(stdout, "\n\nDESCRIPTION: gets a filename, a tag and two words from the\ncommand line and returns the number of occurrences of\nthe tag that contain the two words for that file; for each\noccurrence prints also the opening and closing positions of\nthe tag in the body file.\n\n") ; \
fflush(stdout) ; \
}

#define MACRO_ERROR { \
XCDE_Finalize_Console(&console) ; \
exit(-1) ; \
}

#define MACRO_NO_OCCURRENCES { \
fprintf(stdout, "\n\nThere're no occurrences !!!\n") ; \
XCDE_Finalize_Console(&console) ; \
return(0) ; \
}

// ---------------------------------------------------------------- /
// | this programs reads a filename, a tag and two words from the | / 
// | the command line (in this order) and finds the occurrences   | /
// | of the tag that contains the two words                       | /
// ---------------------------------------------------------------- /

int main(int argc, char *argv[]) 
{ // XCDE console data structure, needed to deal XCDE data. The console must
  // be associated with an XML document that has before been compressed and 
  // indexed using XCDE_Build command. By using the XCDE_Init_Console
  // function the console is associated with an XML document that has been
  // compressed and indexed by the XCDE_Build command. Components of the console
  // are initialized with values regarding that document. All XCDE API functions 
  // get a console reference as a parameter.
  XCDE_Console_Type console ;

  // IL associated with the tag. It's another XCDE data type. An IL is
  // implemented using an array of element of type XCDE_IL_Couple_Type. 
  // This one is a struct with two integer component: XCDE_Position wich
  // contains the position of the occurrence in the body compressed file 
  // (or, better, the position of the codeword in Huffman coding in the
  // body file) and XCDE_Offset, which is different from 0 only when the
  // position associated with the token is derived from the expansion of 
  // an internal entity. With these two values is possible to know if a 
  // word occurrence is contained in a tag occurrence. The IL of a tag has 
  // always a even number of element because its contains the opening and 
  // closing position. 
  XCDE_IL_Couple_Type *tag_IL ;

  // dimension of the tag IL
  int tag_IL_length=0 ;

  // IL associated with the first word .
  XCDE_IL_Couple_Type *word1_IL ;

  // dimension of the first word list of positions.
  int word1_IL_length=0 ; 

  // IL associated with the second word.
  XCDE_IL_Couple_Type *word2_IL ;

  // dimension of the second word list of positions.
  int word2_IL_length=0 ; 

  // this variable is used to contain occurrences of the tag
  // that contain first word occurrences found in a loop. By 
  // using a second loop it's possible to find the occurrences
  // that contain also second word occurrences. These ones will
  // be printed.
  XCDE_IL_Couple_Type *result_tag_IL ;

  // dimension of the results list.
  int result_tag_IL_length=0 ;

  // value returned from a XCDE function: XCDE APIs always return 
  // a value that is 1 is all is correct or -1 otherwise. 
  int res=0 ;

  // rank of the tag in the tags dictionary. The APIs that search in 
  // different dictionaries (TAG, ATTR, WORD) returns the position (or
  // positions, because for attributes and words dictionaries its possibile
  // to return more values) of the element in the specific dictionary. This value
  // (or these values) is called the "rank in the dictionary" and can be a number
  // greater than 0 and less than the total number of elements in the dictionary. 
  // To obtain the IL of the element the rank in the dictionary must be traslated
  // to the general rank, that is the rank in the global dictionary, using a specific
  // API. 
  int tagdictrank=0 ;

  // variable that contains the rank of the tag in the general dictionary
  int taggenrank=0 ;

  // pointer to the array that contains the rank of the first word occurrences in the words dictionary. It's 
  // an array because depending on the type of search it may contain more than one value. 
  int *word1dictrank=NULL ;

  // pointer to the array that contains the rank of the second word occurrences in the words dictionary.
  int *word2dictrank=NULL ;

  // length of the word1dictrank array (must be equal to 1 because the search is EXACT).  
  int numword1indictrank=0 ; 
  
  // length of the word2dictrank array (must be equal to 1 because the search is EXACT).  
  int numword2indictrank=0 ; 

  // the rank of the first word occurrence in the general dictionary (the array must contain only one value because 
  // the search is exact).
  int word1genrank=0 ; 

  // variable that contains the rank of the second word in the general dictionary (the array must contain only one value 
  // because the search is exact).
  int word2genrank=0 ; 

  // variable that contains the pointer to the tag IL (its an integer because IL is implemented as an array).
  int taglistpos=0 ;

  // variable that contains the pointer to words IL. 
  int wordlistpos=0 ;

  // variable that contains the pointer to results IL. 
  int resultlistpos=0 ;

  // number of occurrences in the result list. 
  int num_occ=0 ;

  // checks if the number of parameters is correct.
  MACRO_HEADER

  if (argc == 1) {

    MACRO_HELP
    
    return(1) ;
  }

  if (argc != 5) {

    fprintf(stderr, "\n\nFatal Error: parameters aren't correct\n") ;

    MACRO_HELP

    exit(-1) ; 
  }

  // initializes the console: argv[1] must contain the
  // name of an XML file that has been compressed and indexed 
  // with the XCDE_Build command. If it isn't in the 
  // current directory it must also contain the path. Obviously
  // the source file hasn't to be contained in the directory. 
  // From this moment the console is associated with the 
  // document. 
  res=XCDE_Init_Console(argv[1],
			&console) ;
  
  if (res == -1) MACRO_ERROR 
  
  // argv[2] must contain the name of a tag. This API 
  // checks if the tag is in the tags dictionary and
  // in this case returns its rank in the dictionary.
  // A dictionary rank is always greater than 0.		  
  res=XCDE_Search_Tag_Dictionary(argv[2],
                                 &tagdictrank,
                                 &console) ;

  if (res == -1) MACRO_ERROR

  // checks if there's any occurrences of the tag in the dictionary.
  if (tagdictrank == 0) MACRO_NO_OCCURRENCES 

  // if the program arrives here the tag is contained
  // in the tags dictionary.

  // traslates the rank of the tag to the general rank:
  // it's necessary to access the IL. 
  res=XCDE_Dict2Rank(TAG,
		     tagdictrank,
		     &taggenrank,
		     &console) ;			      
  
  if (res == -1) MACRO_ERROR
		   
  // reads the IL associated to the tag and its length.
  res=XCDE_Get_IL_ListPos(taggenrank,
			  &tag_IL,
			  &tag_IL_length,
			  &console) ;

  // checks if there's any occurrence of the tag in the IL.
  if (tag_IL_length == 0) MACRO_NO_OCCURRENCES 

  // if the program arrives here the IL associated with the 
  // tag isn't empty.

  // argv[3] must contain the name of a word. This API 
  // checks if the word passed is in the words dictionary: it is
  // more complex than the correspondent TAG API, because 
  // different type of search can be made for the words dictionary
  // (for example exact, suffix, prefix, etc.). It's possible
  // that it returns more than one rank, so an array is passed
  // to contain ranks (word1dictrank) and an integer variable 
  // to contain the number of results (num1worddictrank). In this
  // case the search is made exact, so the maximum lenght of the 
  // array is one.  
  res=XCDE_Search_Word_Dictionary(argv[3],
				  &word1dictrank,
				  &numword1indictrank,
				  XCDE_EXACT,
				  XCDE_CASE,
				  0,
				  &console) ; 

  if (res == -1) MACRO_ERROR 

  // checks if there're any occurrences of the first word. If the array length is
  // 0 there're no occurrences. 		   
  if (numword1indictrank == 0) MACRO_NO_OCCURRENCES 

  // argv[4] must contains the name of a word. Checks if it is in the words dictionary: 
  // the search is made with XCDE_EXACT and XCDE_CASE parameters, so the maximum length 
  // of the array is one.				 
  res=XCDE_Search_Word_Dictionary(argv[4],
				  &word2dictrank,
				  &numword2indictrank,
				  XCDE_EXACT,
				  XCDE_CASE,
				  0,
				  &console) ; 

  if (res == -1) MACRO_ERROR 
		   
  // checks if tthe second word is contained in the words dictionary. If the length of the 
  // array is equal to 0 so there're no occurrences.
  if (numword2indictrank == 0) MACRO_NO_OCCURRENCES 				   

  // if the program arrives here the word is contained in the words
  // dictionary.

  // traslates the rank of the first word occurrence to the general rank:
  // it's necessary to access its IL.
  res=XCDE_Dict2Rank(WORD,
		     word1dictrank[0],
		     &word1genrank, 
		     &console) ;
  
  if (res == -1) MACRO_ERROR

  // traslates the rank of the second word occurrence to the general rank:
  // it's necessary to access its IL.		   
  res=XCDE_Dict2Rank(WORD,
		     word2dictrank[0],
		     &word2genrank, 
		     &console) ;
  
  if (res == -1) MACRO_ERROR

  // reads the IL associated to the tag and its length. Remember that the
  // IL is implemented as an array of struct.
  res=XCDE_Get_IL_ListPos(word1genrank,
			  &word1_IL,
			  &word1_IL_length,
			  &console) ;
  
  if (res == -1) MACRO_ERROR

  // checks if the IL associated with the first word is empty. In this 
  // case the token found in the dictionary is physical: it's necessary
  // to compress and decompress the file, but it hasn't been indexed (for
  // example it's in a comment).
  if (word1_IL_length == 0) MACRO_NO_OCCURRENCES
    
  // if the program arrives here the IL associated with the first
  // word isn't empty

  // reads the IL associated to the second word. Remember that the 
  // IL is implemented as an array.
  res=XCDE_Get_IL_ListPos(word2genrank,
			  &word2_IL,
			  &word2_IL_length,
			  &console) ;
  
  if (res == -1) MACRO_ERROR

  // checks if the IL associated with the second word is empty.
  if (word2_IL_length == 0) MACRO_NO_OCCURRENCES

  // if the program arrives here the IL associated with the second
  // word isn't empty.
			      
  // allocates the memory needed to contain the result list
  result_tag_IL=malloc(tag_IL_length*sizeof(XCDE_IL_Couple_Type)) ;

  // loops over the inverted lists to find all tag occurrences of the tag that contain the first word.
  // This happens when word positions are between tag opening and closing positions. In the case there 
  // aren't in a entity expantion it's sufficient to analyze XCDE_Position components of ILs, otherwise
  // (when XCDE_Position components are equal) it's necessary to analyze XCDE_Offset component. 
  while((taglistpos < tag_IL_length) && (wordlistpos < word1_IL_length)) { 
      
      // the word occurrence is in the tag: increases the number of occurrences
      // and the word list pointer.
      if (((tag_IL[taglistpos].XCDE_Position < word1_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos+1].XCDE_Position > word1_IL[wordlistpos].XCDE_Position)) || 
	  
	  ((tag_IL[taglistpos].XCDE_Position == word1_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos].XCDE_Offset < word1_IL[wordlistpos].XCDE_Offset) &&
	   (tag_IL[taglistpos+1].XCDE_Position == word1_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos+1].XCDE_Offset > word1_IL[wordlistpos].XCDE_Offset))) {
	
	result_tag_IL[resultlistpos].XCDE_Position=tag_IL[taglistpos].XCDE_Position ;

	result_tag_IL[resultlistpos].XCDE_Offset=tag_IL[taglistpos].XCDE_Offset ;

	result_tag_IL[resultlistpos+1].XCDE_Position=tag_IL[taglistpos+1].XCDE_Position ;

	result_tag_IL[resultlistpos+1].XCDE_Offset=tag_IL[taglistpos+1].XCDE_Offset ;

	result_tag_IL_length+=2 ;

	resultlistpos+=2 ;

	++wordlistpos ;

	taglistpos+=2 ;
      }
      
      else
	
	// the word occurrence is before the tag: increases the word list pointer.
	if((tag_IL[taglistpos].XCDE_Position > word1_IL[wordlistpos].XCDE_Position) ||
	   ((tag_IL[taglistpos].XCDE_Position == word1_IL[wordlistpos].XCDE_Position) &&
	    (tag_IL[taglistpos].XCDE_Offset > word1_IL[wordlistpos].XCDE_Offset))) 
	  
	  ++wordlistpos ;
      
	else
	  
	  // the word occurrence is after the tag: increases the tag list pointer.
	  if((tag_IL[taglistpos+1].XCDE_Position < word1_IL[wordlistpos].XCDE_Position) ||
	     ((tag_IL[taglistpos+1].XCDE_Position == word1_IL[wordlistpos].XCDE_Position) &&
	      (tag_IL[taglistpos+1].XCDE_Offset < word1_IL[wordlistpos].XCDE_Offset))) 
	    
	    taglistpos+=2 ; 
      
	  else
	    
	    MACRO_ERROR
  }

  resultlistpos=0 ;

  wordlistpos=0 ;

  // loops over the inverted lists to find all tag occurrences that contain not only first
  // word occurrences but also second ones. In this case opening and closing positions of the
  // tag are printed. 
  while((resultlistpos < result_tag_IL_length) && (wordlistpos < word2_IL_length)) { 
      
      // the word occurrence is in the tag: increases the number of occurrences
      // and the word list pointer.
      if (((result_tag_IL[resultlistpos].XCDE_Position < word2_IL[wordlistpos].XCDE_Position) &&
	   (result_tag_IL[resultlistpos+1].XCDE_Position > word2_IL[wordlistpos].XCDE_Position)) || 
	  
	  ((result_tag_IL[resultlistpos].XCDE_Position == word2_IL[wordlistpos].XCDE_Position) &&
	   (result_tag_IL[resultlistpos].XCDE_Offset < word2_IL[wordlistpos].XCDE_Offset) &&
	   (result_tag_IL[resultlistpos+1].XCDE_Position == word2_IL[wordlistpos].XCDE_Position) &&
	   (result_tag_IL[resultlistpos+1].XCDE_Offset > word2_IL[wordlistpos].XCDE_Offset))) {
	
	++num_occ ;

	fprintf(stdout, "\nOccurrence number %d: TAG opening position %d TAG closing position %d", num_occ, result_tag_IL[resultlistpos].XCDE_Position, result_tag_IL[resultlistpos+1].XCDE_Position) ;

	resultlistpos+=2 ;

	++wordlistpos ;
      }
      
      else
	
	// the word occurrence is before the tag: increases the word list pointer.
	if((result_tag_IL[resultlistpos].XCDE_Position > word2_IL[wordlistpos].XCDE_Position) ||
	   ((result_tag_IL[resultlistpos].XCDE_Position == word2_IL[wordlistpos].XCDE_Position) &&
	    (result_tag_IL[resultlistpos].XCDE_Offset > word2_IL[wordlistpos].XCDE_Offset))) 
	  
	  ++wordlistpos ; 
      
	else
	  
	  // the word occurrence is after the tag: increases the tag list pointer.
	  if((result_tag_IL[resultlistpos+1].XCDE_Position < word2_IL[wordlistpos].XCDE_Position) ||
	     ((result_tag_IL[resultlistpos+1].XCDE_Position == word2_IL[wordlistpos].XCDE_Position) &&
	      (result_tag_IL[resultlistpos+1].XCDE_Offset < word2_IL[wordlistpos].XCDE_Offset))) 
	    
	    resultlistpos+=2 ; 
      
	  else
	    
	    MACRO_ERROR
  }

  if (num_occ > 0) { // enters this block if the number of occurrences is greater than 0.

    // prints the result and free the memory allocated.
    fprintf(stdout, "\nNumber of total occurrences: %d\n\n", num_occ) ;

    // frees the memory allocated. The memory is allocated by the APIs but
    // must be freed from the programmer.
    free(word1_IL) ; 
    
    free(word2_IL) ;
    
    free(result_tag_IL) ;
    
    free(tag_IL) ; 
    
    free(word1dictrank) ; 
    
    free(word2dictrank) ;
    
    // finalizes the console. After this operation the memory
    // needed for the console is freed and temporary files 
    // removed. The console variable should be also reused
    // for another document. 
    XCDE_Finalize_Console(&console) ; 
  }

  // in this case there're no occurrences. So prints the message, finalizes the console and exits.
  else MACRO_NO_OCCURRENCES
  
  return(1) ; 
}
     
	  
    
    
