// ********************************************************************
// hashfunc.c
//
// This file contains some auxiliary procedures needed for the
// hash-table manage.
// *********************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hashfunc.h"

extern
FILE *IL_File ;

/* ----------- Create hash table for tokens ----------- */
HashTable
hcreate(unsigned int Table_size)
{ int i;

  HashTable ht;

  ht.Table_size = Table_size ; 

  ht.Table=(Hashptr *) malloc(Table_size*sizeof(Hashptr));

  if (ht.Table == NULL) {
    fprintf(stderr,"Fatal Error: Hash table allocation\n");
    exit(-1);
  }
  
  for (i=0; i<Table_size; i++)
    ht.Table[i]=NULL;
  
  return(ht);
}

/* --------- Compute the hash value for the given token ---------- */
__inline__ int hashfunc(char *s, 
			int len, 
			HashTable ht)
{
  int hfi,hfn;
  
  hfn = 0;
  for (hfi=0; hfi<len ; hfi++)
    hfn = 31 * hfn + (unsigned char) *s++;
  hfn = abs(hfn % ht.Table_size);
  return(hfn);
}

/* ------ Checks wether the hash table contains the given token ------ */
Hashptr hsearch(char *str, 
		int len, 
		unsigned char type, 
		HashTable ht)
{
  Hashptr hsp;

  for (hsp = ht.Table[hashfunc(str,len, ht)]; hsp; hsp = hsp->next)
    if ((len == hsp->len_str) && (strncmp(str,hsp->str,len) == 0) && (hsp->type==type))
	return(hsp);
  return((Hashptr)0);
}

/* --- Insert token in the hash table and returns 1 if new, 0 otherwise --- */
int hinsert(char *str, 
	    int str_len, 
	    unsigned char type, 
	    char *ass_el, 
	    HashTable ht, 
	    int occur, 
	    int escape, 
	    unsigned char par)  
{ register 
    int i=0 ; 
  int hiv;
  Hashptr hip;
  Hashptr table_head;
  Hashptr hip_pred;
  char cont_bit_buf[HASH_CONT_BIT_MAX_LEN] ; 
  unsigned 
    long length_continuation_bit=0 ; 

  hiv=hashfunc(str,
	       str_len, 
	       ht) ;  // compute the hash value

  hip=hsearch(str,
	      str_len,
	      type, 
	      ht );   // check string occurrence

  if (hip != NULL) {
    
    // increases the number of physical occurrences 
    if (CHECK_PARAM(par, 
		    _INCREASE_PHYSICAL_COUNT))
      
      ++hip->count_occ ; 
    
    // compute a logical occurrence 
    if (CHECK_PARAM(par, 
		    _WRITE_LOGICAL_OCCURRENCE)) { 

      continuation_bit_c((unsigned long) (occur-(hip->cmp_rank_list_last_value)+1), 
			 (char *) hip->cmp_rank_list+hip->cmp_rank_list_pointer, 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
      
      hip->cmp_rank_list_pointer+=(length_continuation_bit/8) ;
	      
      hip->cmp_rank_list_last_value=occur;
      
    }
    
    // compute a logical occurrence 
    if (CHECK_PARAM(par, 
		    _INSERT_LOGICAL_OCCURRENCE)) { 
      
      // increases the number of element 
      ++hip->num_rank_list ; 
      
      continuation_bit_c((unsigned long) (occur-(hip->cmp_rank_list_last_value)+1), 
			 (char *) cont_bit_buf, 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
	
      hip->cmp_rank_list_last_value=occur;
      
      hip->cmp_rank_list_byte_length+=(length_continuation_bit/8) ; 
    }

    // insert a physical occurrence in the IL File
    if (CHECK_PARAM(par, 
		    _WRITE_PHYSICAL_OCCURRENCE)) { 

      // the escape sequence eventually opened before is ended 
      if (hip->PosByteEscape >= 0) { 
	
	hip->PosByteEscape=-1 ;
	
	continuation_bit_c((unsigned long) 1,
			   (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 

	hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;
      }
      
      continuation_bit_c((unsigned long) (occur-(hip->cmp_pos_list_last_value)+2),
			 (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 


      hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;

      hip->cmp_pos_list_last_value=occur ;
    }

    // compute a physical occurrence 
    if (CHECK_PARAM(par, 
		    _INSERT_PHYSICAL_OCCURRENCE)) { 
      
      // the escape sequence eventually opened before is ended 
      if (hip->PosByteEscape >= 0) { 
	
	hip->PosByteEscape=-1 ;

	continuation_bit_c((unsigned long) 1,
			   (char *) cont_bit_buf,
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 

	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ;
      }
	
      // increases the number of elements
      ++hip->num_byte_list ; 
      
      continuation_bit_c((unsigned long) occur-(hip->cmp_pos_list_last_value)+2,
			 (char *) cont_bit_buf,
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
      
      hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ;	
      
      hip->cmp_pos_list_last_value=occur ;
    }
    
    if (CHECK_PARAM(par,
		    _WRITE_PHYSIC_ESCAPE)) { 
      
      if (hip->PosByteEscape < 0) { 
	
	hip->PosByteEscape=occur ; 
	
	continuation_bit_c((unsigned long) 1,
			   (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;	

	// adds the value 2 to the difference to code the -1 element 
	// and to avoid the zero problems 
	continuation_bit_c((unsigned long) occur-(hip->cmp_pos_list_last_value)+2,
			   (char *) hip->cmp_pos_list+hip->cmp_pos_list_pointer, 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 

	hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;	
	
	hip->cmp_pos_list_last_value=occur; 
      }
      
      else
	
	if ((hip->PosByteEscape >= 0) && (hip->PosByteEscape != occur)) { 
	  
	  hip->PosByteEscape=occur ; 
	  
	  continuation_bit_c((unsigned long) 1,
			     (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			     (unsigned int) 0,
			     (unsigned long *) &length_continuation_bit) ; 
	  
	  hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;	

	  continuation_bit_c((unsigned long) occur-(hip->cmp_pos_list_last_value)+2,
			     (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			     (unsigned int) 0,
			     (unsigned long *) &length_continuation_bit) ; 

	  hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;	
	  
	  hip->cmp_pos_list_last_value=occur; 
	}
      
      continuation_bit_c((unsigned long) escape+2,
			 (char *) (hip->cmp_pos_list+hip->cmp_pos_list_pointer), 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
      
      hip->cmp_pos_list_pointer+=(length_continuation_bit/8) ;		
    }
    
    if (CHECK_PARAM(par,
		    _INSERT_PHYSIC_ESCAPE)) { 
      
      if (hip->PosByteEscape < 0) { 
	
	hip->PosByteEscape=occur ; 
	
	continuation_bit_c((unsigned long) 1,
			   (char *) cont_bit_buf,
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ;	
	
	// adds the value 2 to the difference to code the -1 element 
	// and to avoid the zero problems 
	continuation_bit_c((unsigned long) occur-(hip->cmp_pos_list_last_value)+2,
			   (char *) cont_bit_buf, 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ; 

	hip->cmp_pos_list_last_value=occur; 
      }
      
      else
	
	if ((hip->PosByteEscape >= 0) && (hip->PosByteEscape != occur)) { 

	  hip->PosByteEscape=occur ; 
	
	  continuation_bit_c((unsigned long) 1,
			   (char *) cont_bit_buf,
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	  
	  hip->cmp_pos_list_byte_length+=((length_continuation_bit/8)*2) ;	
	  
	  continuation_bit_c((unsigned long) occur-(hip->cmp_pos_list_last_value)+2,
			     (char *) cont_bit_buf, 
			     (unsigned int) 0,
			     (unsigned long *) &length_continuation_bit) ; 
	
	  hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ; 

	  hip->cmp_pos_list_last_value=occur; 
	}
      
      // increases the number of elements in the positions list
      ++hip->num_byte_list ; 
      
      continuation_bit_c((unsigned long) escape+2,
			 (char *) cont_bit_buf, 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
	
      hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ; 
    }

    // move-to-front the searched token
    if (ht.Table[hiv] != hip){
      
      // Find hip predecessor which surely does exist
      // because hip is not the head of the list
      for (hip_pred = ht.Table[hiv]; hip_pred->next != hip;
	   hip_pred = hip_pred->next) ;
      
      hip_pred->next = hip->next;   // jump hip in the old list
      table_head = ht.Table[hiv];   // save the old head of the list-entry
      ht.Table[hiv] = hip;        // hip is the new head of the list
      hip->next = table_head;
    }
    
    return(0);
  }
  
  else {    //------ The token is new -------
    
    hip = (Hashptr) malloc(sizeof(Hashnode));
    
    if (hip == NULL){
      fprintf(stderr,"Fatal Error: Insert hash table\n");
      exit(-1); }
    
    if (CHECK_PARAM(par, 
		    _INSERT_ASSOCIATED_ELEMENT))
      
      hip->ass_el = ass_el ; 
    
    else hip->ass_el = NULL ; 
    
    hip->len_str = str_len ;
    
    hip->type = type ; 

    hip->str = strdup(str) ; 
    
    hip->str[str_len]='\0' ;
    
    hip->PosByteEscape=-1 ; 
    
    hip->num_rank_list=0 ; 
    
    hip->num_byte_list=0 ; 
    
    hip->count_occ=1 ;
    
    hip->tag_number_open_flag=0 ;

    hip->cmp_rank_list_pointer=0 ;
    
    hip->cmp_pos_list_pointer=0 ;
    
    hip->cmp_rank_list_byte_length=0 ; 
    
    hip->cmp_rank_list=NULL ; 
    
    hip->cmp_pos_list=NULL ; 

    hip->cmp_pos_list_byte_length=0 ; 

    hip->cmp_rank_list_last_value=0 ;

    hip->cmp_pos_list_last_value=0 ; 

    hip->next=ht.Table[hiv] ;
    
    ht.Table[hiv]=hip ;
    
    if (CHECK_PARAM(par,
		    _INCREASE_PHYSICAL_COUNT)) 
      
      ++hip->count_occ ; 
    
    if (CHECK_PARAM(par,
		    _INSERT_LOGICAL_OCCURRENCE)) { 

      // increases the number of element 
      ++hip->num_rank_list ; 
      
      continuation_bit_c((unsigned long) (occur-(hip->cmp_rank_list_last_value)+2), 
			 (char *) cont_bit_buf, 
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
      
      hip->cmp_rank_list_last_value=occur;
      
      hip->cmp_rank_list_byte_length+=(length_continuation_bit/8) ; 
    }
    
    if (CHECK_PARAM(par,
		    _INSERT_PHYSICAL_OCCURRENCE)) { 

      // increases the number of elements
      ++hip->num_byte_list ; 
      
      continuation_bit_c((unsigned long) (occur-(hip->cmp_pos_list_last_value)+2),
			 (char *) cont_bit_buf,
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ; 
      
      hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ;	
      
      hip->cmp_pos_list_last_value=occur ;
    }

    if (CHECK_PARAM(par,
		    _INSERT_PHYSIC_ESCAPE)) { 

      if (hip->PosByteEscape < 0) { 

	hip->PosByteEscape=occur ; 

	continuation_bit_c((unsigned long) 1,
			   (char *) cont_bit_buf,
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ;	
	
	// adds the value 2 to the difference to code the -1 element 
	// and to avoid the zero problems 
	continuation_bit_c((unsigned long) (occur+2),
			   (char *) cont_bit_buf, 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ; 

	hip->cmp_pos_list_last_value=occur; 

	++hip->num_byte_list ; 

	continuation_bit_c((unsigned long) escape+2,
			   (char *) cont_bit_buf, 
			   (unsigned int) 0,
			   (unsigned long *) &length_continuation_bit) ; 
	
	hip->cmp_pos_list_byte_length+=(length_continuation_bit/8) ; 
      }
    }
    return(1);
  }
}

void hdestroy(HashTable ht)
{ register int k=0 ; 

 Hashptr p ;

 Hashptr q ; 

 for(k=0 ; k<ht.Table_size; ++k) 

   if (ht.Table[k] != NULL) { 

     p=ht.Table[k] ; 
     
     while (p != NULL) { 
       
       q=p->next ;
       
       free(p) ;
       
       p=q ; 
     }
   }

 ht.Table=NULL ;
 
 ht.Table_size=0 ; 

 return ; 
}
 
int hopentag(Hashptr tag) 
{
  ++tag->tag_number_open_flag ;

  return(tag->tag_number_open_flag) ;
}

int hclosetag(Hashptr tag)
{ 
  --tag->tag_number_open_flag ;

  return(tag->tag_number_open_flag) ; 
}
