/********************************************************************
  	         	   xcde_huffman.c
                         contains functions: 
         Sort_for_freq, Sort_for_cwlen, Huffman_byte_code, 
                  Creates_HT_leaves, Build_Hufftree, 
         Compute_codeword_lengths, output_token, nooutput_token
	            Xml Compressed Documents Engine
	                   Building Program
	        Andrea Mastroianni - Paolo Ferragina 
          Dipartimento di Informatica - Universita' di Pisa 
	               ferragin@di.unipi.it
                      mastroi@cli.di.unipi.it
********************************************************************/
#include"xcde_consts.h"
#include"xcde_build.h"

extern
FILE *inputFile ; 

extern
FILE *configFile ; 

extern
FILE *dictsFile[XCDE_NUMBER_DICT] ; 

extern
FILE *bodyFile ;

extern
FILE *headerFile ; 

extern
FILE *IL_File ; 

extern
FILE *tagstructFile ; 

extern
FILE *debugPhysicView ; 

extern
FILE *debugIndexFile ; 

extern
FILE *debugStatsFile ; 

extern
char *inputFilename ; 

extern
char *dictionaryFilename[XCDE_NUMBER_DICT] ; 

extern
char *headerFilename ;

extern
char *bodyFilename ; 

extern
char *IL_Filename ; 

extern
char *tagstructFilename ; 

extern
char *debugPhysicFilename ; 

extern
char *debugStatsFilename ; 

extern
char *debugIndexFilename ; 

extern
char *configFilename ; 

extern
int inputFilesize ; 

extern
char zlib_decompress_buffer[XCDE_COMPRESS_BLOCK_SIZE] ; 

extern
char zlib_compress_buffer[2*XCDE_COMPRESS_BLOCK_SIZE] ; 

extern
int pointer_decompress_buffer ; 

extern
unsigned
char verbosing_flag ; 

extern
int number_tokens ; 

extern
HashTable ht ; 

extern 
char *expat_source_pointer ; 

extern
char *xml_code ; 

extern
long xml_dimension ; 

extern
long xml_position ; 

extern
unsigned 
char changed_position ; 

extern
char *body_pointers_cmp_buffer ; 

extern
int body_pcmp_point ; 

extern
int body_pointers_cmp_buffer_dimension ; 

extern
int number_body_points ; 

extern
int diff_position ; 

extern
int new_diff_position ; 

extern
int new_body_pointers_point ; 

extern
int body_size_without_zlib ; 

extern
HashTable ht ; 

extern
HashTable expat_general_declared_entity ;

extern
HashTable expat_separators_table ; 

extern
HashTable expat_attribute_separators_table ; 

extern
Hashptr *Huffman_tree ; 
  

/*******************************************************************
                       sorts for tokens frequencies
 *******************************************************************/

int Sort_for_freq(const void *va, 
		  const void *vb) 
{
  return(((Hashptr *) vb)[0]->count_occ - ((Hashptr *) va)[0]->count_occ);
}

/*******************************************************************
                        sorts for codeword length
 *******************************************************************/

int Sort_for_cwlen(const void *va, 
		   const void *vb) 
{  
  if ((((Hashptr *) vb)[0]->cw_len - ((Hashptr *) va)[0]->cw_len) != 0)
    return((((Hashptr *) vb)[0]->cw_len - ((Hashptr *) va)[0]->cw_len));

  return(((Hashptr *) va)[0]->type - ((Hashptr *) vb)[0]->type) ;

  if(((Hashptr *) vb)[0]->str == NULL)
    return(-1);

  if(((Hashptr *) va)[0]->str == NULL)
    return(1);

  return(strcmp(((Hashptr *) vb)[0]->str,((Hashptr *) va)[0]->str));
}

/*******************************************************************
          construct Huffman code using a tree with fan-out 128
 *******************************************************************/

Hashptr *Huffman_byte_code(int *firstcw, 
			   int *offsetcw, 
			   int *count_depth)
{ register
    int i=0 ;

  Hashptr p=NULL ;

  Hashptr *Tree_array=NULL ;

  // maxim codeword length
  int max_cwlen=5 ;

  // Huffman tree root
  int Huff_root=0 ;

  int HT_size=0 ;

  int HT_leaves=0 ;

  int HT_dummy_leaves=0 ;

  int HT_nondummy_leaves=0 ;
  
  HT_dummy_leaves = 128 - (number_tokens % 127);

  HT_nondummy_leaves = number_tokens ; 

  HT_leaves = HT_nondummy_leaves + HT_dummy_leaves;

  while (max_cwlen > 4) { 

    if((Tree_array=(Hashptr *)malloc(sizeof(Hashptr)*2*HT_leaves))==NULL) { 

      fprintf(stderr,"\n\tFatal Error: out of memory during tree_array creation\n");
      
      DELETE_FILES ; 
      
      exit(-1);  
    }
    
    // creates leaves sorted by decreasing frequencies 
    Create_HT_leaves(Tree_array,
		     HT_leaves);

    // builds the Huffman tree 
    Huff_root = Build_HuffTree(Tree_array,
			       &HT_size,
			       HT_leaves);
      
    if (verbosing_flag) { 

      fprintf(stdout, "\n\nHuffman tree dimension %d\n", HT_size) ; 
      
      fprintf(stdout, "Not dummy leaves %d\n", HT_nondummy_leaves) ; 
      
      fprintf(stdout, "Dummy leaves %d\n", HT_dummy_leaves) ; 
    }
    
    max_cwlen = Compute_codeword_lengths(Tree_array,
					 Huff_root,
					 HT_size);
    
    if(max_cwlen > 4) { 

      fprintf(stdout, "\n\t ----  Codeword length bigger than 4  ---- \n") ; 
      
      for(i=0;i<HT_nondummy_leaves;i++) { 

	p = Tree_array[i+Huff_root]; 
	
	// scaling using golden ratio
	p->count_occ = p->count_occ / 1.618 + 1; 
      }
      
      free(Tree_array);
    }
  }
  
  if (verbosing_flag)
    
    fprintf(stdout, "\nMaximum codeword length %d", max_cwlen) ; 
  
  // computes first codewords and offsets 
  for(i=4;i>0;i--) { 

    count_depth[i] = 0;
    
    offsetcw[i] = 0;
    
    firstcw[i] = 0;
  }
  
  for(i=0;i<HT_nondummy_leaves;i++)
    
    count_depth[Tree_array[i+Huff_root]->cw_len]++;
  
  offsetcw[max_cwlen] = 0;
  
  firstcw[max_cwlen] = HT_dummy_leaves;  
  
  for(i=max_cwlen - 1;i>0;i--) { 

    offsetcw[i] = offsetcw[i+1]+count_depth[i+1];
    
    firstcw[i] = (firstcw[i+1]+count_depth[i+1]) >> 7;
  }
  
  firstcw[max_cwlen] = 0;  
  
  for(i=4;i>0;i--)
    
    if (count_depth[i] == 0) 
      
      offsetcw[i] = HT_leaves+1;
  
  if (verbosing_flag) { 

    fprintf(stdout, "\n---Offset value %d means there aren't codewords of that length", HT_leaves+1) ; 
    
    for(i=4; i>0;i--)
      printf("\nOffset[%d] = %d, First_cw[%d] = %d, count_cw[%d] = %d\n",
	     i,offsetcw[i],i,firstcw[i],i,count_depth[i]);
  }

  return(Tree_array+Huff_root);
}

/*******************************************************************
              creates leaves sorted for decreasing frequencies 
 *******************************************************************/

void Create_HT_leaves(Hashptr *tree, 
		      int HT_leaves)
{ // variables used as counters
  register 
    int i=0 ; 

  register 
    int j=0 ; 

  Hashptr p=NULL ;

  for(i=0,j=0; i<number_tokens;j++) { 

    for(p = ht.Table[j]; p ; p = p->next)
      tree[i++] = p;
  }
  
  for(i=number_tokens; i<HT_leaves; i++) { 

    if ((tree[i] = (Hashptr) malloc(sizeof(Hashnode))) == NULL) { 

      fprintf(stderr,"\n\tFatal Error: Token parsing\n");
      
      DELETE_FILES ; 
      
      exit(-1); 
    } 
    
    else { 

      tree[i]->str = NULL;  

      tree[i]->len_str=0 ;

      tree[i]->count_occ = 0;
    }
  }
  
  qsort(tree,
	HT_leaves,
	sizeof(Hashptr),
	Sort_for_freq);
  
  return ; 
}

/*******************************************************************
                  creates Huffman canonical tree 
 *******************************************************************/

int Build_HuffTree(Hashptr *work_area, 
		   int *tot_nodes, 
		   int HT_leaves)
{ // counter variables 
  register 
    int i=0 ; 

  register
    int count=0 ;

  Hashptr p=NULL;

  int num_nodes=0 ; 

  int processed=0 ;

  // INVARIANT:
  // work_area[0...num_nodes]
  //        contains nodes which parents hasn't been created yet
  // work_area[num_nodes+1...processed]
  //        not used, but contain changes dinamically 
  // work_area[processed+1...2*HT_leaves-1]
  //        contains nodes which parents hasn't been created 

  processed = 2 * HT_leaves - 1;
 
  for(num_nodes = HT_leaves-1; num_nodes > 0; ) { 

    // gets the smallest 128 elements as new element children 
    for(i=0,count=0; i<128; i++) { 

      // copy_node(work_area,num_nodes,processed);
      work_area[processed] = work_area[num_nodes];
      
      count+=work_area[num_nodes]->count_occ;
      
      processed--;
      
      num_nodes--;
    }
    
    for(i=num_nodes; (i>=0) && (work_area[i]->count_occ < count); i--)
      
      work_area[i+1]=work_area[i]; 
    
    num_nodes++;  
    
    // new parent node 
    if ((p = (Hashptr) malloc(sizeof(Hashnode))) == NULL) { 

      fprintf(stderr,"\n\tFatal Error: Token parsing\n");
      
      DELETE_FILES ; 
      
      exit(-1); 
    }
    
    // values of the internal node 
    p->str = NULL; 

    p->count_occ = count;

    p->codeword = 0;
    
    p->cw_len = 0;
    
    p->next = (Hashptr) (processed + 1); // tick
    
    work_area[i+1]=p;
  }
  
  // Copy_node(work_area,0,processed);
  work_area[processed] = work_area[0];  
  
  *tot_nodes = 2 * HT_leaves - processed;
  
  return(processed);
}

/*******************************************************************
        computes codeword lengths and dictionaries composition
 *******************************************************************/

int Compute_codeword_lengths(Hashptr *tree, 
			     int root, 
			     int tree_size)
{ register
    int i=0 ; 

  register 
    int j=0 ; 

  Hashptr *p=NULL ; 
 
  int offset_child=0 ;

  int isleaf=0 ;
 
  int max_length=0 ;


  // top-down tree exploration 
  for(i=0; i<tree_size; i++) { 

    offset_child = (int) (tree[root+i]->next);  
    
    isleaf = (tree[root+i]->str != NULL) || (tree[root+i]->count_occ == 0);  
    
    if (isleaf == 0) { 

      for(j=0; j<128; j++) { 

	if(tree[offset_child+j]->count_occ == 0) { 

	  tree[offset_child+j]->cw_len = -1; 	
	} 
	
	else 
	  { tree[offset_child+j]->cw_len = (tree[root+i]->cw_len) + 1;
	  }
	
	if(tree[offset_child+j]->cw_len > max_length)
	  
	  max_length = tree[offset_child+j]->cw_len;
      }
      
      // for the next sorting step 
      tree[root+i]->cw_len = -1;  
    }
  }
  
  for(i=0; i<tree_size; i++) { 

    offset_child = (int) (tree[root+i]->next);
    
    isleaf = (tree[root+i]->str != NULL) || (tree[root+i]->count_occ == 0);  
  }
  
  qsort(tree+root,
	tree_size,
	sizeof(Hashptr),
	Sort_for_cwlen);
  
  return max_length;
}

/*******************************************************************
                    puts codewords into the body file 
 *******************************************************************/

void output_token(char *token, 
		  int token_len,
		  unsigned char type)
{ 
  register 
    int j=0 ;

  register
    int i=0 ; 
 
  unsigned 
    long len_continuation_bit ;

  Hashptr p=NULL ;

  // ----------------------
  // VARIABLES USED BY LZIP
  // ----------------------

  Byte *uncompr ; 

  uLong comprLen ;

  uLong uncomprLen ; 
  
  int cmp_err=0 ; 

  // checks if the token is in the Hash table 
  if((p = hsearch(token,
		  token_len, 
		  type, 
		  ht)) == NULL){ 

    fprintf(stderr,"\n\tFatal Error: writing error (output_token) %s\n", token) ; 
    
    DELETE_FILES ; 
    
    exit(-1); 
  }    
  
  // byte with leading 1
  zlib_decompress_buffer[pointer_decompress_buffer]=((p->codeword >> (7 * (p->cw_len - 1))) | 0x80) ; 
  
  ++pointer_decompress_buffer ; 
  
  if (pointer_decompress_buffer == XCDE_COMPRESS_BLOCK_SIZE) { 

    if (diff_position > 0) { 

      // checks if there is enough space in the buffer to compress block point
      if ((body_pcmp_point) > (body_pointers_cmp_buffer_dimension-10)) { 

	body_pointers_cmp_buffer=realloc(body_pointers_cmp_buffer, 
					 body_pointers_cmp_buffer_dimension*2) ; 
	
	body_pointers_cmp_buffer_dimension*=2 ; 
      }
      
      continuation_bit_c((unsigned long) new_diff_position, 
			 (char *) body_pointers_cmp_buffer+body_pcmp_point,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation_bit) ; 
      
      body_pcmp_point+=(len_continuation_bit/8) ;
	
      ++number_body_points ; 
    }
    
    uncompr=(Byte *) zlib_decompress_buffer ; 
    
    uncomprLen=pointer_decompress_buffer ; 
    
    comprLen=2*XCDE_COMPRESS_BLOCK_SIZE ; 
    
    cmp_err=compress2((Bytef *) zlib_compress_buffer,
		      (uLongf *) &comprLen, 
		      (const Bytef *) uncompr, 
		      (uLong) uncomprLen, 
		      Z_BEST_COMPRESSION) ;
    
    if (cmp_err != Z_OK) { 

      fprintf(stderr, "\n\tFatal Error: problems with zlib compression\n") ; 
      
      DELETE_FILES ; 
      
      exit(-1) ; 
    }
    
    for (i=0 ; i<comprLen; ++i) 
      
      fputc((char) zlib_compress_buffer[i], 
	    bodyFile) ; 
    
    diff_position+=comprLen ; 
    
    new_diff_position=comprLen ; 
    
    pointer_decompress_buffer=0 ; 
  }
  
  for(j=(p->cw_len-2); j>=0; j--){ 

    zlib_decompress_buffer[pointer_decompress_buffer]=((p->codeword >> (7 * j)) & 0x7F) ; 
    
    ++pointer_decompress_buffer ;
    
    if (pointer_decompress_buffer==XCDE_COMPRESS_BLOCK_SIZE) { 

      if (diff_position > 0) { 

	// checks if there is enough space in the buffer to compress block point
	if ((body_pcmp_point) > (body_pointers_cmp_buffer_dimension-10)) { 

	  body_pointers_cmp_buffer=realloc(body_pointers_cmp_buffer, 
					   body_pointers_cmp_buffer_dimension*2) ; 
	  
	  body_pointers_cmp_buffer_dimension*=2 ; 
	}
	
	continuation_bit_c((unsigned long) new_diff_position, 
			   (char *) body_pointers_cmp_buffer+body_pcmp_point,
			   (unsigned int) 0,
			   (unsigned long *) &len_continuation_bit) ; 
	
	body_pcmp_point+=(len_continuation_bit/8) ;
	
	++number_body_points ; 
      }
      
      uncompr=(Byte *) zlib_decompress_buffer ; 
      
      uncomprLen=pointer_decompress_buffer ; 
      
      comprLen=2*XCDE_COMPRESS_BLOCK_SIZE ; 
      
      cmp_err=compress2((Bytef *) zlib_compress_buffer,
			(uLongf *) &comprLen, 
			(const Bytef *) uncompr, 
			(uLong) uncomprLen, 
			Z_BEST_COMPRESSION) ;
      
      if (cmp_err != Z_OK) {

	fprintf(stderr, "\n\tFatal Error: problems with zlib compression\n") ; 
	
	DELETE_FILES ; 
	
	exit(-1) ; 
      }
      
      for (i=0 ; i<comprLen; ++i) 
	
	fputc((char) zlib_compress_buffer[i], 
	      bodyFile) ; 
      
      diff_position+=comprLen ; 
      
      new_diff_position=comprLen ; 
      
      pointer_decompress_buffer=0 ; 
    }
  }
  
  body_size_without_zlib+=p->cw_len ; 
  
  return ; 
}

/*******************************************************************
       update body size dimension by adding codeword length
 *******************************************************************/

void nooutput_token(char *token, 
		    int token_len,
		    unsigned char type)
{ Hashptr p=NULL ;

  // checks if the token is in the Hash table 
  if((p = hsearch(token,
		  token_len, 
		  type, 
		  ht)) == NULL) { 

    fprintf(stderr,"\n\tFatal Error: writing error (output_token) %s\n", token) ; 
    
    DELETE_FILES ; 
    
    exit(-1); 
  }    
  
  body_size_without_zlib+=p->cw_len ; 
}
