/********************************************************************
		         xcde_tagstruct.c 
	          Xml Compressed Document Engine
	                 Building Program
		Paolo Ferragina - Andrea Mastroianni
           Dipartimento di Informatica - Universita' di Pisa 
	               ferragin@di.unipi.it
                       mastroia@di.unipi.it
********************************************************************/

// *****************************************************************
// This software has been derived from the Filtering Data Structure
// of Chazelle [SICOMP 96] which allows the execution of time and space 
// optimal stabbing queries.
//
// Author: Paolo Ferragina
// Date: December 2001
// *****************************************************************

#include "stdlib.h"
#include "stdio.h"
#include "math.h"
#include <unistd.h>
#include <sys/mman.h>
#include "xcde_consts.h"
#include "xcde_build.h"

#define DELTA 1000   // Controls time vs. space trade-off

// ======================= TYPES ==========================================
typedef struct {
  uint phis_pos;
  uint displ;
} pos;

typedef struct {
  pos position;
  uint tag_pos_dict;
} int_endpoint;
// ============ AUXILIARY FUNCTIONS  ====================================
void error_signal(char *s)
{
  printf("\n\n %s\n",s);
  exit(1);
}

// Comparison function for interval sorting
int cmp_intervalli(const void *a, 
		   const void *b)
{
  int_endpoint *aa, *bb;
  uint a_val, b_val;
  int cmp;
  
  aa = (int_endpoint *) a;
  bb = (int_endpoint *) b;
  cmp = (aa->position.phis_pos == bb->position.phis_pos);
  a_val = (cmp ? aa->position.displ : aa->position.phis_pos);
  b_val = (cmp ? bb->position.displ : bb->position.phis_pos);

  return ( a_val - b_val ); 
}

// Copies a position from B to A
void copy_pos(pos *A, 
	      pos B)
{

  A->phis_pos=B.phis_pos;
  A->displ=B.displ;
}

// Copies an endpoint from B[b] to A[a]  
void copy_interval(int_endpoint *A, 
		   int_endpoint *B, 
		   int a, 
		   int b)
{
  copy_pos(&(A[a].position),B[b].position);
  A[a].tag_pos_dict=B[b].tag_pos_dict;
}

// Check if an endpoint is an opening (Pila on triples)
int open_tag(int_endpoint *Pila, 
	     int size, 
	     int_endpoint *A, 
	     int i)
{
  if (size==0) return(1);

  if (Pila[size-1].tag_pos_dict != A[i].tag_pos_dict) return(1);

  return(0);
}

// Check if a tag-id is an opening  (Pila on tag ids)
int open_tag2(int *Pila,
	      int size, 
	      int tag)
{
  if (size==0) return(1);

  if (Pila[size-1] != tag) return(1);

  return(0);
}

// ================== BUILDING FUNCTION  ====================================
//
// This is the procedure of Chazelle properly modified to host
// our data structures.
//
// A contains the endpoints (phis_pos, displ, tag_id) of input
// AA keeps the endpoints subdivided among windows.
// P keeps the indices of AA where each window starts.
// S keeps the (phis,displ) position where a window starts [included].
//
int Build_Windows(int_endpoint *A, 
		  int_endpoint *AA, 
		  int *P, 
		  pos *S, 
		  int n)
{ int i=0 ; 
  int j=0 ;
  int r=0 ; 
  int h=0 ; 
  int low=0 ; 
  int cur=0 ; 
  int T=0 ;
  
  int dimension_Pila=5000 ; 

  int_endpoint *Pila ;

  int Pila_size=0 ;

  pos start_win ;

  
  // allocates the memory needed by the Pila
  Pila=malloc(dimension_Pila*sizeof(int_endpoint)) ; 

  P[0]=0;  // First window starts at the first endpoint, obviously
  S[0].phis_pos=0; // Starting position of the first window
  S[0].displ=0;

  T=0;     // #intervals currently in the window
  cur=0;   // number of currently stabbed intervals
  low=1;   // min # of stabbed intervals in the current window

  j=0;  // moves over P 
  r=0;  // moves over AA


  for(i=0; i<n; i++) { 

    // We cycle over all the endpoints
    
    if (open_tag(Pila,
		 Pila_size,
		 A,
		 i)) { 

      // Endpoint denoting an open-tag
      
      copy_pos(&start_win,
	       A[i].position); // Record its position
      
      cur++; // new stabbed interval
      
      T++; // new interval in the window
      
      if ((DELTA * low) < T) { 

	// Window is too full
	
	j++; // Initializes a new window
	
	P[j]=r; // Starting index in AA of the new window
	
	copy_pos(S+j,
		 start_win); // S[j] = start pos of the new window
	
	
	// Copy the endpoints of the intervals yet open
	for(h=0; h<Pila_size; h++) { 

	  copy_interval(AA,
			Pila,
			r,
			h);
	  r++; 
	}
	
	copy_interval(AA,
		      A,
		      r,
		      i);  // copy (open) endpoint into the window
	
	r++;	
	
	copy_interval(Pila,
		      A,
		      Pila_size,
		      i); // copy it into the Pila
	
	Pila_size++;
	
	// reallocates the Pila 
	if(Pila_size >= dimension_Pila) { 

	  dimension_Pila*=2 ; 
	  
	  Pila=realloc(Pila, 
		       dimension_Pila) ; 
	}
	
	
	T=cur; // Update current # intervals in the window	
	
	low = T; // Update current stabbing number
	
      } 
      
      else 
	
	{  // Window is not yet full (yet opening endpoint)

	  copy_interval(AA,
			A,
			r,
			i); // Insert endpoint into the window
	  
	  r++; 

	  copy_interval(Pila,
			A,
			Pila_size,
			i); // Insert it in the Pila
	  
	  Pila_size++;
	  
	  // reallocates the Pila 
	  if(Pila_size >= dimension_Pila) {  

	    dimension_Pila*=2 ; 
	    
	    Pila=realloc(Pila, 
			 dimension_Pila) ; 
	  }
	  
	}
      
    } else 
      
      {  // Case of a closing endpoint
	copy_interval(AA,
		      A,
		      r,
		      i);   // Copy the closing endpoint
	
	r++;	

	Pila_size--;  // Remove the interval from the Pila
	
	// Record the ending position of the window
	copy_pos(&start_win,
		 A[i].position); 
	
	start_win.displ++;  // Increment the littlest
	
	// T is not updated because accounts for the # intervals
	
	    cur--;  // Update # stabbed intervals
	    
	    low = ((low < cur) ? low : cur); // update lowest # stabbed
	    
	    if ((DELTA * low) < T) { // If the window is full (case closing)
	      
	      j++;  // Initializes a new window
	      
	      P[j]=r; // Record the starting index in AA
	      
	      copy_pos(S+j,
		       start_win); // Record starting position (special one)
	      
	      // Copy the intervals yet open
	      for(h=0; h<Pila_size; h++) {

		copy_interval(AA,
			      Pila,
			      r,
			      h);
		
		r++; 
	      }
	      
	      // Now we update T because we are in a new window
	      T = cur;  // cur remains unchanged because it is correct
	      
	      low = ((1 < T) ? T : 1);
	    }
      }
  }
  
  // Surely writes an empty window at the end.
  //  S[j] is the first uncovered point 
  // P[j] keeps the total number of endpoints we wrote
  return(j);  // returns the number of windows we built
}

// ====================== I/Os FUNCTION  ====================================

// win_array contains the list of (position,tag), one window after the other
// win_pointers cointains the starting element of each window into win_array
// Virtual_File contains infos about the windows compressed via DELTA-code
// Virtual_StartWin contains the starting bytes of each window
int Write_Virtual_File(int num_windows, 
		       int_endpoint *win_array, 
		       int *win_pointers,
		       uchar *Virtual_File, 
		       uint *Virtual_StartWin)
{ int i=0 ; 

  int j=0 ; 
  
  int r=0 ; 
  
  int num_int=0 ; 

  uint diff;

  uchar *Virtual_pointer;

  int pointer=0 ; 

  uchar *uncmp_buffer ; 

  int uncmp_buffer_dimension=200000 ; 

  int comprLen=0 ; 

  int uncomprLen=0 ; 

  int cmp_err=0 ; 

  unsigned 
    long len_continuation_bit=0ul ;

  Virtual_pointer = Virtual_File;

  uncmp_buffer=(char *) malloc(uncmp_buffer_dimension*sizeof(char)) ; 

  // Scans through the windows: r global index
  for(r=0,i=0;i<num_windows;i++) { 

    // Starting byte of the window
    Virtual_StartWin[i]= (uint) ((int)Virtual_pointer - (int)Virtual_File);
    
    // Scans through the "num_int" intervals of a window
    num_int = win_pointers[i+1]-win_pointers[i];
    
    if (uncmp_buffer_dimension < ((num_int*3)*2*(sizeof(int)+1))) { 

      free(uncmp_buffer) ; 
      
      uncmp_buffer_dimension=((num_int*3)*2*(sizeof(int)+1)) ; 
      
      uncmp_buffer=(char *) calloc(uncmp_buffer_dimension,
				   sizeof(char)) ; 
    }
    
    pointer=0 ;
    
    for(j=0; j<num_int; j++)  { 

      if (j==0) 
	
	{ 
	  diff = 1 + win_array[r].position.phis_pos;
	} 
      
      else 
	
	{ 
	  diff = 1 + (win_array[r].position.phis_pos - 
		      win_array[r-1].position.phis_pos);
	}
      
      continuation_bit_c((unsigned long) diff, 
			 (char *) uncmp_buffer+pointer,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation_bit) ;
      
      pointer+=(len_continuation_bit/8) ;
      
      
      continuation_bit_c((unsigned long) win_array[r].position.displ+1, 
			 (char *) uncmp_buffer+pointer,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation_bit) ;
      
      pointer+=(len_continuation_bit/8) ;
      
      continuation_bit_c((unsigned long) win_array[r].tag_pos_dict+1, 
			 (char *) uncmp_buffer+pointer,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation_bit) ;
      
      pointer+=(len_continuation_bit/8) ;
      
      r++;
    }
    
    uncomprLen=pointer ;
    
    comprLen=uncmp_buffer_dimension ; 
    
    // compresses the buffer
    cmp_err=compress2((Bytef *) Virtual_pointer , 
		      (uLongf *) &comprLen,
		      (const Bytef *) uncmp_buffer , 
		      (uLong) uncomprLen, 
		      Z_BEST_COMPRESSION) ;
    
    if (cmp_err != Z_OK)
      
      { fprintf(stderr, "\nProblems with compression") ; 
      
      exit(-1) ; 
      }
    
    Virtual_pointer+=comprLen ; 
  }
  
  // frees memory
  free(uncmp_buffer) ; 
  
  return((int)Virtual_pointer - (int)Virtual_File);
}

// Write in a file ".context" the windows with an HEADER
//
// Header contains:
// 1. #windows (4 bytes)
// 2. for each window
//        #triples (phis_pos, displ, tag) into it 
//        the starting byte in the Context_file where the window is coded
//        the starting position (phisical, displacement) of the window
//
// The body of the Context_File contains, for each window:
//     the triples coded using delta-code (on phis_pos we store the difference)
//
// We have chosen to sum 1 to every number to avoid the (possible) coding of 0.
// 
void Write_File(FILE *tagstructfile, 
		int num_windows, 
		int *Window_pointers, 
		uchar *Virtual_File, 
		uint *Virtual_StartWin, 
		pos *Start_points, 
		int Virtual_File_len)
{ 
  char cont_buf[XCDE_CONT_BIT_BUFFER_MAX_LENGTH] ;

  unsigned long len_cont=0ul ;  

  int tmp, i, k;
 
  continuation_bit_c((unsigned long) num_windows,
		     (char *) cont_buf, 
		     (unsigned int) 0,
		     (unsigned long *) &len_cont) ; 	
  
  // scrittura della lunghezza delle finestre 
  for(k=0 ; k<(len_cont/8); ++k)
    
    fputc(cont_buf[k],
	  tagstructfile) ; 
  
  // Header: 4 bytes per 4 elements
  for(i=0; i<num_windows; i++) { 

    // window size = # triples into it
    tmp=Window_pointers[i+1]-Window_pointers[i];
    
    continuation_bit_c((unsigned long) tmp,
		       (char *) cont_buf, 
		       (unsigned int) 0,
		       (unsigned long *) &len_cont) ; 	
    
    for(k=0 ; k<(len_cont/8); ++k)
      
      fputc(cont_buf[k],
	    tagstructfile) ; 
    
    // Virtual Byte start (without header) ; 
    continuation_bit_c((unsigned long) Virtual_StartWin[i],
		       (char *) cont_buf, 
		       (unsigned int) 0,
		       (unsigned long *) &len_cont) ; 
    
    
    for(k=0 ; k<(len_cont/8); ++k)
      
      fputc(cont_buf[k],
	    tagstructfile) ; 
    
    // text position of start (phisical and displacement) 
    continuation_bit_c((unsigned long) Start_points[i].phis_pos, 
		       (char *) cont_buf, 
		       (unsigned int) 0,
		       (unsigned long *) &len_cont) ; 
    
    
    for(k=0 ; k<(len_cont/8); ++k)
      
      fputc(cont_buf[k],
	    tagstructfile) ; 
    
    // text position of start (phisical and displacement) 
      continuation_bit_c((unsigned long) Start_points[i].displ,
			 (char *) cont_buf, 
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ; 
      
      
      for(k=0 ; k<(len_cont/8); ++k)
	
	fputc(cont_buf[k],
	      tagstructfile) ; 
  }
  
  // Write the body of the .tagstruct file
  fwrite(Virtual_File,
	 sizeof(uchar),
	 Virtual_File_len,
	 tagstructfile) ; 
}

// ================== BUILDING PROCEDURES ===============================
void Build_Context_File(FILE *tagstructfile,
			Hashptr *Huffman_tree,
			int ntokens,
			int num_intervals)
{ int_endpoint *int_array ;   // Array of 2n endpoints

 int_endpoint *win_array ;   // Array of windows

 int *win_pointers;         // Array of ranks for identifying the windows
 
 uint *Virtual_StartWin ;
 
 uchar *Virtual_File;
 
 pos *Start_points ;     // Array of points where window start
 
 int tot_size_win=0 ; 
 
 int num_windows;
 
 int tmp;
 
 int VF_length;
 
 int pointer=0 ; 

 int elem=0 ; 

 int read_el=0 ;

 unsigned 
   long len=0l ;

 uchar escape=0 ; 
 
 int i=0 ;

 int int_array_position=0 ; 

 int_array = (int_endpoint *) calloc((2*num_intervals), sizeof(int_endpoint));

 for(i=0 ; i<ntokens; ++i)

   if (Huffman_tree[i]->type == TAG) { 

     escape=0 ; 
     
     pointer=0 ; 
     
     elem=0 ; 
     
     while(pointer < Huffman_tree[i]->cmp_pos_list_byte_length) {
       
       continuation_bit_d((unsigned long *) &read_el, 
			  (char *) (Huffman_tree[i]->cmp_pos_list+pointer), 
			  (unsigned int) 0,
			  (unsigned long *) &len) ;
       
       pointer+=(len/8) ; 
       
       if ((read_el == 1) && (!escape)) { 
	 
	 escape=1 ; 
	 
	 continuation_bit_d((unsigned long *) &read_el, 
			    (char *) (Huffman_tree[i]->cmp_pos_list+pointer), 
			    (unsigned int) 0,
			    (unsigned long *) &len) ;
	 
	 pointer+=(len/8) ; 
       }
       
       else
	 
	 if ((read_el == 1) && (escape)) 
	   
	   escape=0 ;
       
	 else
	   
	   if ((read_el != 1) && (escape)) {  
	     
	     int_array[int_array_position].position.phis_pos=elem ; 
	     
	     int_array[int_array_position].position.displ=(read_el-2) ;   
	     
	     int_array[int_array_position].tag_pos_dict=i ; 
	     
	     ++int_array_position ;
	   }
       
	   else { 
	     
	     elem+=(read_el-2) ;
	     
	     int_array[int_array_position].position.phis_pos=elem ; 
	     
	     int_array[int_array_position].position.displ=0 ;  
	     
	     int_array[int_array_position].tag_pos_dict=i ; 
	     
	     ++int_array_position ;  
	   }
     }
     
     if (Huffman_tree[i]->cmp_pos_list != NULL) { 

       free(Huffman_tree[i]->cmp_pos_list) ;
       
       Huffman_tree[i]->cmp_pos_list=NULL ; 
       
       Huffman_tree[i]->cmp_pos_list_pointer=0 ;
       
       Huffman_tree[i]->cmp_pos_list_last_value=0 ; 
     }
   }
 
 tot_size_win = (int) ((4 * (DELTA) * num_intervals) / ((DELTA) - 1) + 5) ;
 
 win_array = (int_endpoint *) malloc(tot_size_win*sizeof(int_endpoint)) ;
 
 win_pointers = (int *) malloc((size_t) tot_size_win*sizeof(int)) ; 
 
 Start_points = (pos *) malloc(tot_size_win*sizeof(pos)) ;
 
 // Sort the endpoints
 qsort(int_array,            
       2*num_intervals, 
       sizeof(int_endpoint),
       cmp_intervalli) ;    
 
 // win_pointers[num_windows]=#elements in all windows
 num_windows=Build_Windows(int_array, 
			   win_array, 
			   win_pointers,
			   Start_points, 
			   2*num_intervals) ;
 
 if (win_pointers[num_windows]>=tot_size_win)
   
   error_signal("ERROR: total size windows is too large\n") ;

 free(int_array) ;
 
 // Uses a virtual space
 tmp = (int) ceil(1.2 * tot_size_win * sizeof(int_endpoint)) ;
 
 Virtual_File = (uchar *) malloc(tmp) ;
 
 clean_buffer(Virtual_File,
	       tmp);
 
 Virtual_StartWin = (uint *) malloc ((num_windows+1) * sizeof(uint));
 
 // Compresses the windows in the Virtual File
  VF_length=Write_Virtual_File(num_windows,
			       win_array,
			       win_pointers,
			       Virtual_File,
			       Virtual_StartWin);
  
  // Write the file for the context data structure
  Write_File(tagstructfile, 
	     num_windows,
	     win_pointers,
	     Virtual_File,
	     Virtual_StartWin,
	     Start_points,
	     VF_length);

  // Cleans the memory
  free(Virtual_StartWin);

  free(Virtual_File);

  free(win_pointers);

  free(Start_points);
}

