 /********************************************************************
			       xcde_lib.h
                   header file needed by the XCDE API library 
	                Xml Compressed Document Engine
	             Paolo Ferragina - Andrea Mastroianni 
             Dipartimento di Informatica - Universita' di Pisa 
	                   ferragin@di.unipi.it
	                  mastroi@cli.di.unipi.it
 ********************************************************************/
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h> 
#include<sys/mman.h> 
#include<sys/stat.h>
#include<string.h>
#include<ctype.h>
#include<unistd.h>
#include"limits.h"
#include"xcde_consts.h"
#include"zlib.h"
#include"hashfunc.h"

#define uchar unsigned char 

// structure that contains only physical 
// positions of the IL  
typedef
struct XCDE_IL_Couple_Struct_Type
{ 
  int XCDE_Position ; 

  int XCDE_Offset ; 
} XCDE_IL_Couple_Type ; 

// structure that contains IL physical positions
// and offset 
typedef 
struct XCDE_IL_Triple_Struct_Type
{ 
  int XCDE_Position ; 

  int XCDE_Offset ; 

  int XCDE_Proximity ;

} XCDE_IL_Triple_Type ; 

// structure used to contain information 
// in the tagstruct header
typedef
struct XCDE_Tagstruct_Windows_Information_Type_Struct
{ 
  int XCDE_Tagstruct_Window_Size ; 

  int XCDE_Tagstruct_Window_Byte_Displacement ; 

  int XCDE_Tagstruct_Window_Start ; 

  int XCDE_Tagstruct_Window_Displ ; 

} XCDE_Tagstruct_Windows_Information_Type ;

/*******************************************************************
	                   XCDE LIB CONSOLE 
 *******************************************************************/ 

typedef 
struct XCDE_Console_Type_Struct 
{ unsigned char XCDE_Console_Initialized ; 

  // flag for the proximity counters presence  
  unsigned char XCDE_Proximity ; 

  // flag for the tagstruct presence 
  unsigned char XCDE_Tagstruct ; 

  /* ----------- */
  /* file source */
  /* ----------- */

  int XCDE_XML_Source_Filesize ; 

  char *XCDE_XML_Source_File_Name ; 

  /* ---------------- */ 
  /* dictionary files */ 
  /* ---------------- */

  // pointers to temporary files that contain the decompressed 
  // dictionaries 
  FILE *XCDE_Dict_Dcmptmp_File_Pointers[XCDE_NUMBER_DICT] ; 

  char *XCDE_Dict_Cmp_File_Names[XCDE_NUMBER_DICT] ; 
  
  // flag for the dictionary decompression 
  unsigned char XCDE_Dict_File_Decompressed[XCDE_NUMBER_DICT] ; 

  // floag for the dictionary loading 
  unsigned char XCDE_Dict_File_Loaded[XCDE_NUMBER_DICT] ; 

  char *XCDE_Dict_Dcmptmp_Names[XCDE_NUMBER_DICT] ;

  char *XCDE_Dict_Dcmptmp_Mapped_Pointers[XCDE_NUMBER_DICT] ; 

  int XCDE_Dict_Dcmptmp_Filesize[XCDE_NUMBER_DICT] ; 

  /* ----------- */ 
  /* file header */ 
  /* ----------- */ 

  FILE *XCDE_Header_File_Pointer ; 

  char *XCDE_Header_File_Name ; 

  char *XCDE_Header_Mapped_Pointer ;

  int XCDE_Header_Filesize ; 

  /* ------- */ 
  /* file IL */ 
  /* ------- */ 

  FILE *XCDE_IL_File_Pointer ; 

  char *XCDE_IL_File_Name ; 

  char *XCDE_IL_Mapped_Pointer ; 

  int XCDE_IL_Filesize ; 

  // table for IL accessing 
  int *XCDE_IL_Table_Pointers ; 
  
  int XCDE_IL_Table_Pointers_Size ; 

  /* --------- */ 
  /* file body */
  /* --------- */ 

  FILE *XCDE_Body_File_Pointer ;

  char *XCDE_Body_File_Name ; 

  char *XCDE_Body_Mapped_Pointer ; 

  int XCDE_Body_Num_Block ; 

  // table for body zlib block accessing 
  int *XCDE_Body_Table_Pointers ; 

  // zlib cache block loaded in the memory 
  char *XCDE_Body_Cache_Block ; 

  int XCDE_Body_Cache_Which_Block ; 

  int XCDE_Body_Cache_Dcmp_Block_Size ; 

  int XCDE_Body_Cache_Cmp_Block_Size ; 

  int XCDE_Body_Cache_Block_File_Position ; 

  int XCDE_Body_Filesize ; 

  int XCDE_Body_Size_Without_Zlib ; 

  /* -------------- */ 
  /* file tagstruct */
  /* -------------- */ 

  FILE *XCDE_Tagstruct_File_Pointer ; 

  int XCDE_Tagstruct_Filesize ; 

  char *XCDE_Tagstruct_File_Name ; 

  char *XCDE_Tagstruct_Mapped_Pointer ;

  // tagstruct window decompressed in the memory 
  char *XCDE_Tagstruct_Cache_Window ; 

  int XCDE_Tagstruct_Cache_Which_Block ; 

  int XCDE_Tagstruct_Number_Windows ; 

  int XCDE_Tagstruct_Cache_Dcmp_Block_Size ; 

  int XCDE_Tagstruct_Cache_Cmp_Block_Size ; 
  
  // header structure for tagstruct accessing 
  XCDE_Tagstruct_Windows_Information_Type *XCDE_Tagstruct_Windows_Information ; 

  /* ------------------------------------ */ 
  /* Huffman tree and dictionary division */ 
  /* ------------------------------------ */ 

  int Huffman_Codeword_Max_Len ; 

  int Huffman_Offsetcw[XCDE_HUFF_MAX_LEVEL] ;

  int Huffman_Firstcw[XCDE_HUFF_MAX_LEVEL];

  int Huffman_Num_Tokens ; 

  // general dictionary division 
  int Huffman_Dict_NumEl[XCDE_HUFF_MAX_LEVEL+1][XCDE_NUMBER_DICT] ; 

  // number of elements in a single level 
  int Huff_NumEl_Lev[XCDE_HUFF_MAX_LEVEL+1] ; 
  
  // pointers to dictionary tokens 
  char **Huffman_Position_Token ; 

} XCDE_Console_Type ; 

/*******************************************************************
                    General APIs to access the console				     
*******************************************************************/ 

// function that initializes the console data structure.
// Example : ret=XCDE_Init_Console("example.xml", &Console) ; 
// where "example.xml" is the name of the source file and 
// console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Init_Console(const char *name_source,
		      XCDE_Console_Type *XCDE_Console) ;


// function that finalizes the console data structure and
// frees the memory allocated by the XCDE_Init_Console function. 
// Example : ret=XCDE_Finalize_Console(&Console) ; 
// where &Console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Finalize_Console(XCDE_Console_Type *XCDE_Console) ; 

// function that reads the proximity flag.
// Example : ret=XCDE_SettedProximity(&prox, &Console) ; 
// where &prox is the memory address of the variable that will 
// contain the proximity flag and &Console is the memory address of 
// the console data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_SettedProximity(unsigned char *ProxFlag,
			 XCDE_Console_Type *XCDE_Console) ; 

// functions that reads the tagstruct flag, that is on if the
// tagstruct file is present, and of otherwise. 
// Example : ret = XCDE_SettedTagstruct(&TagstructFlag, &Console) 
// where &TagstructFlag is the memory address of the variable that
// will contain the tagstruct flag, and &Console is the memory 
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_SettedTagstruct(unsigned char *TagstructFlag,
			 XCDE_Console_Type *XCDE_Console) ; 

// function that reads the initialized console flag.
// Example : ret=XCDE_ConsoleInitialized(&initflag, &Console) ; 
// where initflag is the memory address of the variable that will
// contain the console initialized flag and &Console is the memory 
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_ConsoleInitialized(unsigned char *InitializeFlag, 
			    XCDE_Console_Type *XCDE_Console) ; 

/*******************************************************************
		   APIs to read file source informations
 *******************************************************************/ 

// function that reads the size of the XML source file.
// Example : ret=XCDE_Get_Source_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain the 
// size, and &Console is the memory address of the console data structure 

int XCDE_Get_Source_Filesize(int *Source_Filesize,
			     XCDE_Console_Type *XCDE_Console) ;

// function that reads the name of the XML source file.
// Example : ret=XCDE_Get_Source_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Source_Filename(char **Source_Filename, 
			     XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		    APIs to read file source informations 
 *******************************************************************/ 

// function that reads the size of the header file.
// Example : ret=XCDE_Get_Header_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain the 
// size, and &Console is the memory address of the console data structure 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Header_Filesize(int *Header_Filesize,
			     XCDE_Console_Type *XCDE_Console) ;

// function that reads the name of the header file.
// Example : ret=XCDE_Get_Header_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Header_Filename(char **Header_Filename, 
			     XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		    APIs to read file body informations 
 *******************************************************************/ 

// function that reads the size of the the body file 
// Example : ret=XCDE_Get_Body_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain
// the size and &Console is the memory address of the console 
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Filesize(int *Body_Filesize,
			   XCDE_Console_Type *XCDE_Console) ;

// function that reads the name of the body file.
// Example : ret=XCDE_Get_Body_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Filename(char **Body_Filename, 
			   XCDE_Console_Type *XCDE_Console) ;

// function that puts in Eof_Flag a flag which indicates
// if the position Body_Position is after the end of the Body,
// and so doesn't indicate a valid codeword. 
// Example : ret=XCDE_Body_Eof(body_position, &eofflag, &Console) 
// where body_position is the position we want to check, &eofflag is
// the memory address of the flag and &Console is the memory address
// of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Body_Eof(int Body_Position,
		  unsigned char *Eof_Flag, 
		  XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		       APIs to access file body 
 *******************************************************************/ 

// function that reads the codeword that begins at the specified position.
// Example : ret=XCDE_Get_Body_Forward_Codeword(position, &Cw, &Cwlen, &Console)
// where position is the position where the reading begins, &Cw is the memory
// address of the variable that will contain the codeword, &Cwlen is the
// memory address of the variable that will contain the codeword length and 
// &Console is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Forward_Codeword(int Position,
				   int *Codeword,
				   int *Cwlen,
				   XCDE_Console_Type *XCDE_Console) ;

// function that reads the previous codeword at the specified position.
// Example : ret=XCDE_Get_Body_Backward_Codeword(position, &Cw, &Cwlen, &Console)
// where position is the specified position, &Cw is the memory address of the
// variable that will contain the codeword, &Cwlen is the memory address of the
// variable that will contain the codeword length and &Console is the memory
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Backward_Codeword(int Position,
				    int *Codeword,
				    int *Cwlen,
				    XCDE_Console_Type *XCDE_Console) ;

// function that reads a text window. The reading begins at position and
// goes forward and backward until the function passed as parameter is 
// true Count times backward and forward.
// Example : ret=XCDE_Get_Body_TextWindow(position, &Window, &func, &Console) ;
// where position is the position the extraction begin, &Window is the 
// memory address where the window is copied, &func is the memory address of
// the function used to check the codeword (gets as parameters the codeword 
// read and its length), Count is the number of times the function must be 
// true.
// The function allocates the memory for the window.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_TextWindow(int Position, 
			     char **Window,
			     int Count, 
			     unsigned char direction,
			     int *backward_new_position,
			     int *forward_new_position,
			     int (*Counter_function)(int cw, int cwlen, XCDE_Console_Type *XCDE_Console),
			     XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		    APIs to read IL file informations 
 *******************************************************************/ 

// function that reads the size of the the IL file 
// Example : ret=XCDE_Get_IL_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain
// the size and &Console is the memory address of the console 
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_Filesize(int *IL_Filesize,
			 XCDE_Console_Type *XCDE_Console) ;

// function that reads the name of the IL file.
// Example : ret=XCDE_Get_IL_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_Filename(char **IL_Filename, 
			 XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		       APIs to access IL file 
 *******************************************************************/ 

// function that gets the IL of positions of the token with the specified Rank.
// Example : ret=XCDE_Get_IL_ListPos(Rank, &IL, &IL_Num_El, &Console)
// where Rank is the rank of the element, &IL is the memory address of
// the variable that will contain the IL, &IL_Num_El is the memory address of
// the variable that will contain the number of elements of the IL and &Console
// is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_ListPos(int Rank,
			XCDE_IL_Couple_Type **IL,
			int *IL_Num_El,
			XCDE_Console_Type *XCDE_Console) ;

// function that reads the IL whit positions and proximities of the token with 
// the specified Rank.
// Example : ret=XCDE_Get_IL_ListPosProx(Rank, &IL, &IL_Num_El, &Console)
// where Rank is the rank of the element, &IL is the memory address of
// the variable that will contain the IL, &IL_Num_El is the memory address of
// the variable that will contain the number of elements of the IL and &Console
// is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_ListPosProx(int Rank,
			    XCDE_IL_Triple_Type **IL,
			    int *IL_Num_El,
			    XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
	    APIs to read dictionary files informations 
 *******************************************************************/ 

// function that reads the name of the specified compressed dictionary 
// file.
// Example : ret=XCDE_Get_Dict_Cmp_Filename(TAG, &name, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &name is the memory address of the variable 
// that will contain the name and &Console is the memory address of 
// the console data structure.
// The function allocates the memory for the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Cmp_Filename(unsigned char Dict,
			       char **Dict_Cmp_Filename,
			       XCDE_Console_Type *XCDE_Console) ;

// function that gets the size of the specified compressed dictionary 
// file.
// Example : ret=XCDE_Get_Dict_Cmp_Filesize(TAG, &size, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &size is the memory address of the variable 
// that will contain the size and &Console is the memory address of 
// the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Cmp_Filesize(unsigned char Dict,
			       int *Dict_Cmp_Filesize,
			       XCDE_Console_Type *XCDE_Console) ; 

// function that reads the name of the temporary file where has been 
// decompressed the specified dictionary file.
// Example : ret=XCDE_Get_Dict_Dcmptmp_Filename(TAG, &name, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &name is the memory address of the variable 
// that will contain the name and &Console is the memory address of 
// the console data structure.
// The function allocates the memory for the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Dcmptmp_Filename(unsigned char Dict,
				   char **Dict_Dcmptmp_Filename,
				   XCDE_Console_Type *XCDE_Console) ;

// function that gets the size of the temporary file where has been 
// decompressed the specified dictionary file.
// Example : ret=XCDE_Get_Dict_Dcmptmp_Filesize(TAG, &size, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &size is the memory address of the variable 
// that will contain the size and &Console is the memory address of 
// the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Dcmptmp_Filesize(unsigned char Dict,
				   int *Dict_Dcmptmp_Filesize, 
				   XCDE_Console_Type *XCDE_Console) ;


/*******************************************************************
		  APIs to access dictionary files 
 *******************************************************************/ 

// function that loads the tokens of the specified dictionary 
// (that must be decompressed, otherwise returns -1) in the
// memory. 
// Example : ret=XCDE_Load_Dictionary(TAG, &Console) 
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR) and Console is the memory address of the
// console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Load_Dictionary(unsigned char Dict,
			 XCDE_Console_Type *XCDE_Console);

// function that decompress the specified dictionary if it hasn't been
// decompressed yet (in this case returns -1).
// Example : ret=XCDE_Decompress_Dictionary(TAG, &Console) 
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR) and &Console is the memory address 
// of the console data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Decompress_Dictionary(unsigned char Dict,
			       XCDE_Console_Type *XCDE_Console) ;

// function that looks for tokens in the word dictionary and
// gets the numbers of the dictionary element where the each element
// is.
// Example : ret=XCDE_Search_Word_Dictionary(token, &token, &num_token, XCDE_PREFIX, XCDE_CAPUND, 0, &Console) 
// where token is the pointer to the string to look for, &token is the 
// memory address of the array  will contain elements in the dictionary, 
// &num_token is the memory address of the variable that will contain the number of elements,
// XCDE_PREFIX is the type of search (may be XCDE_SUFFIX, XCDE_EXACT, XCDE_CONTAINED, XCDE_ERROR, XCDE_REGEXP),
// XCDE_CASE means that there's difference between capital and undercase letters (may be otherwise XCDE_NOCASE) in the search,
// 0 is the number of errors (may be 1 or 2, and can be used only with XCDE_ERROR) and Console is the memory address of the console data structure. 

int XCDE_Search_Word_Dictionary(char *String_To_Search,
				int **TokenDict,
				int *NumToken,
				unsigned char type, 
				unsigned char lettercase, 
				unsigned int nerror, 
				XCDE_Console_Type *XCDE_Console) ;

int XCDE_Search_Tag_Dictionary(char *String_To_Search,
			       int *Token,
			       XCDE_Console_Type *XCDE_Console) ;

int XCDE_Search_Attr_Dictionary(char *attributename,
				char *attributevalue, 
				int **TokenDict,
				int *NumToken,
				unsigned char typeattributesearch,
				unsigned char typevaluesearch, 
				unsigned char lettercaseattribute, 
				unsigned char lettercasevalue, 
				unsigned int nerrorattribute,
				unsigned int nerrorvalue, 
				XCDE_Console_Type *XCDE_Console) ;

// function that reads the number of tokens that are in the general 
// dictionary.
// Example : ret=XCDE_Get_Num_Tokens(&num_tokens, &Console) 
// where &num_tokens is the memory address of the variable that will contain 
// the number of tokens and &Console is the memory address of the console
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_Tokens(int *Num_Tokens,
			XCDE_Console_Type *XCDE_Console) ;

// function that reads the number of tokens that are in the specified dictionary.
// Example : ret=XCDE_Num_El_Dict(TAG, &NumTokens, &Console) 
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &NumTokens is the memory address of the variable 
// that will contain the number of tokens and &Console is the memory 
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_El_Dict(unsigned char Dict,
			 int *Num_El,
			 XCDE_Console_Type *XCDE_Console) ;

// function that reads the number of tokens in a specified 
// dictionary wich codeword length is lev
// Example : ret=XCDE_Num_El_Dictlev(lev, TAG, &NumTokens, &Console) 
// where lev is the level,  TAG (WORD, ATTR) is the constant for 
// the dictionary of TAG (WORD, ATTR), &NumTokens is the memory 
// address of the variable that will contain the number of tokens 
// and &Console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_El_Dictlev(int Lev,
			    unsigned char Dict,
			    int *Num_El,
			    XCDE_Console_Type *XCDE_Console) ;  

// function that reads the maximum codeword length of the Huffman code.
// Example : ret=XCDE_Get_Maxcwlen(&Maxcwlength, &Console)
// where &Maxcwlenght is the memory address of the variable that
// will contain the maximum codeword length and &Console is the memory
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Maxcwlen(int *Maxcwlen,
		      XCDE_Console_Type *XCDE_Console) ;

// function that gets the number of tokens which codeword 
// lenght is lev.
// Example : ret=XCDE_Num_El_Lev(lev, &NumTokens, &Console) 
// where lev is the level, &NumTokens is the memory address of 
// the variable that will contain the number of tokens and &Console 
// is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_El_Lev(int Lev,
			int *Num_El,
			XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
	       APIs to read tagstruct file informations 
 *******************************************************************/ 

// functions that reads the size of the tagstruct file (returns
// -1 if itsn't present).
// Example : ret=XCDE_Get_Tagstruct_Filesize(&Size, &Console) 
// where &size is the memory address of the variable that will
// contain the tagstruct file size, and &Console is the memory 
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Filesize(int *Tagstruct_Filesize,
				XCDE_Console_Type *XCDE_Console) ;

// functions that reads the tagstruct file name (returns -1 
// if theresn't the tagstruct file). 
// Example : ret=XCDE_Get_Tagstruct_Filename(&TagstructFilename, &Console)
// where &TagstructFilename is the memory address of the variable that 
// will contain the file name, and &Console is the memory address of 
// the console data structure.
// The function allocates the memory needed by the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Filename(char **Tagstruct_Filename, 
				XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
		  APIs to access the tagstruct
 *******************************************************************/ 

// functions that reads the tagstruct flag, that is on if the
// tagstruct file is present, and of otherwise. 
// Example : ret = XCDE_SettedTagstruct(&TagstructFlag, &Console) 
// where &TagstructFlag is the memory address of the variable that
// will contain the tagstruct flag, and &Console is the memory 
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_SettedTagstruct(unsigned char *TagstructFlag,
			 XCDE_Console_Type *XCDE_Console) ;

// function that accesses the tagstruct file and computes the tags 
// that contain the position specified (return -1 is theresn't the 
// tagstruct file).
// Example : ret=XCDE_Get_Tagstruct_Tags_Enclosing(&point, &Tags, &Num_tags, &Console) 
// where point is the position (couple physical position, offset), &Tags is the
// memory address of the variable that will contain the array of tags, 
// &Num_tags is the numbers of tags that are in the array and &Console is the
// memory address of the console data struct.
// The function allocates the memory needed by the array of tags 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Tags_Enclosing(XCDE_IL_Couple_Type *Stab_point,
				      int **Enclosing_tags_ranks,
				      int *Num_enclosing_tags,
				      XCDE_Console_Type *XCDE_Console) ;

/*******************************************************************
     APIs to traslate the data obtained from different structures 
 *******************************************************************/ 

// functions that traslates the position of a token in a specified 
// dictionary to its codeword length. 
// Example : ret=XCDE_Get_DictEl2Lev(TAG, position, &length, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR), position if the position of the element in
// the specified dictionary, &length is the memory address of the 
// variable that will contain the lenght, and &Console is the
// memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_DictEl2Lev(unsigned char Dict,
		    int NDictElem,
		    int *Lev,
		    XCDE_Console_Type *XCDE_Console) ;

// functions that traslates the position of a token in a specified 
// dictionary to its rank in the general dictionary (the dictionary 
// of all the tokens).
// Example : ret=XCDE_Get_Dict2Rank(TAG, position, &rank, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR), position if the position of the element in
// the specified dictionary, &rank is the memory address of the 
// variable that will contain the rank computed and &Console is the
// memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Dict2Rank(unsigned char Dict,
		   int NDictElem,
		   int *Rank,
		   XCDE_Console_Type *XCDE_Console) ; 

// function that translates the rank of a token in the general 
// dictionary to the position in the dictionary where it's in. 
// Example : XCDE_Rank2Dict(rank, &Dict, &DictPos, &Console)
// where rank is the rank of the token in the general dictionary,
// &Dict is the memory address of the variable that will contain
// the dictionary, &DictPos is the memory address of the variable
// that will contain the position and &Console is the memory
// address of the console data struct
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Rank2Dict(int Rank,
		   unsigned char *Dict,
		   int *NDictElem,
		   XCDE_Console_Type *XCDE_Console) ;

// function that traslates the codeword to its rank.
// Example : ret=XCDE_Codeword2Rank(cw, cwlen, &Rank, &Console)
// where cw is the codeword, cwlen is its length, &Rank is the 
// memory address of the variable that will contain the Rank and
// &Console is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Codeword2Rank(int Codeword,
		       int Cwlen, 
		       int *Rank,
		       XCDE_Console_Type *XCDE_Console) ;

// function that traslates the codeword to the token corresponding it. 
// Example : ret=XCDE_Codeword2Token(cw, cwlen, &Token, &Console) 
// where cw is the codeword to traslate, cwlen is its length, &Token
// is the memory address of the variable that will contain the token
// and &Console is the memory address of the console data struct.
// The function allocates the memory for the token.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Codeword2Token(int Codeword,
			int Cwlen,
			char **Token,
			XCDE_Console_Type *XCDE_Console) ;

// function that traslates a rank to the token corresponding it. 
// Example : ret=XCDE_Rank2Token(rank, &Token, &Console) 
// where rank is the rank of the token, &Token is the memory address 
// of the variable that will contain it and &Console is the memory 
// address of the console data struct.  
// The function allocates the memory for the token.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Rank2Token(int Rank,
		    char **Token,
		    XCDE_Console_Type *XCDE_Console) ;

