/********************************************************************
		          xcde_search.c 
	         Xml Compressed Documents Engine
		         Search Program
		 Paolo Ferragina - Andrea Mastroianni
         Dipartimento di Informatica - Universita' di Pisa 
	               ferragin@di.unipi.it
                      mastroi@cli.di.unipi.it
********************************************************************/
#include"xcde_search_language.h"
#include"xcde_lib.h"
#include<unistd.h>
#include"sys/time.h"

#define ATTRIBUTE_ERROR                                                                                         \
{ fprintf(stderr, "%sAttribute option of tag number %d is not valid\n", XCDE_ERROR_PARAMETERS, tagcounter) ;    \
  fflush(stderr) ;                                                                                              \
  if (IL_Tag != NULL)                                                                                           \
     free(IL_Tag) ;                                                                                             \
  if (IL_Tags != NULL)                                                                                          \
     free(IL_Tags) ;                                                                                            \
	(stderr) ;                                                                                              \
  XCDE_Finalize_Console(&XCDE_Console) ;                                                                        \
  exit(-1) ;                                                                                                    \
}

#define WORD_ERROR                                                                                 \
{ fprintf(stderr, "%sWord option number %d is not valid\n", XCDE_ERROR_PARAMETERS, wordcounter) ;  \
  fflush(stderr) ;	 	                                                                   \
  XCDE_Finalize_Console(&XCDE_Console) ;                                                           \
  exit(-1) ;                                                                                       \
}
 

// function used to join lists of positions of attributes or words and
// created a new sorted list that contains all the position   
int join_attrs_or_words(XCDE_IL_Couple_Type **IL_new, 
			int *IL_new_num_el,
			XCDE_IL_Couple_Type *IL1,
			int IL1_num_el,
			XCDE_IL_Couple_Type *IL2,
			int IL2_num_el) ;

// function used to join lists of positions of words 
int join_words(XCDE_IL_Triple_Type **IL_new, 
	       int *IL_new_num_el,
	       XCDE_IL_Triple_Type *IL1,
	       int IL1_num_el,
	       XCDE_IL_Triple_Type *IL2,
	       int IL2_num_el) ; 

// function used to select all occurrencies of a tag that doesn't contain 
// one or more attributes 
int select_tag_without_attrs(XCDE_IL_Couple_Type **selected_IL,
			     int *num_el_selected_IL,
			     XCDE_IL_Couple_Type *tag_IL,
			     int num_el_tag_IL,
			     XCDE_IL_Couple_Type *attr_IL,
			     int num_el_attr_IL) ;

// functions used to select all occurrences of tags that contains one 
// or more attributes
int select_tag_with_attrs(XCDE_IL_Couple_Type **selected_IL,
			  int *num_el_selected_IL,
			  XCDE_IL_Couple_Type *tag_IL,
			  int num_el_tag_IL,
			  XCDE_IL_Couple_Type *attr_IL,
			  int num_el_attr_IL) ;

// function used to select all occurrencies of a tag that 
// are nested in other tag
int select_tag_nested_in_tag(XCDE_IL_Couple_Type **selected_IL,
			     int *num_el_selected_IL,
			     XCDE_IL_Couple_Type *tag_IL1,
			     int num_el_tag_IL1,
			     XCDE_IL_Couple_Type *tag_IL2,
			     int num_el_tag_IL2) ;

// function used to select all occurrencies of a tag that
// are nested in other tag with the specified distance
int select_tag_nested_in_tag_with_distance(XCDE_IL_Couple_Type **selected_IL,
					   int *num_el_selected_IL,
					   int rank, 
					   int dist,
					   XCDE_IL_Couple_Type *tag_IL1,
					   int num_el_tag_IL1,
					   XCDE_IL_Couple_Type *tag_IL2,
					   int num_el_tag_IL2,
					   XCDE_Console_Type *Console) ;

// function used to select all occurencies of a word
// that are contained in a tag 
int select_words_contained_in_tags(XCDE_IL_Triple_Type **selected_IL, 
				   int *num_el_selected_IL, 
				   XCDE_IL_Couple_Type *tag_IL, 
				   int num_el_tag_IL, 
				   XCDE_IL_Triple_Type *word_IL,
				   int num_el_word_IL) ; 

// function used to select all minimal windows results that're contained
// in the tag list
int select_proximity_windows_contained_in_tags(XCDE_IL_Couple_Type **selected_IL, 
					       int *selected_IL_num, 
					       XCDE_IL_Couple_Type *tag_IL,
					       int num_el_tag_IL, 
					       XCDE_IL_Couple_Type *windows_results, 
					       int num_el_windows_results) ; 

// function used to compute the proximity windows with the plane
// sweep algorithm
int plane_sweep_proximity(int number_proximal_lists, 
			  XCDE_IL_Triple_Type **Proximal_lists,
			  int *Proximal_Lists_Length,
			  XCDE_IL_Couple_Type **windows_result,
			  int *number_windows_result,
			  int proximity) ;

int main(int argc,
	 char *argv[])
{ register
    int k=0 ;

 register
   int j=0 ;

 register
   int i=0 ;

 unsigned
   char tagstruct_flag=0 ; 

 unsigned
   char proximity_flag=0 ; 

 int numarg=0 ; 

 int res=0 ;
  
 char *tokentosearch=NULL ;

 int tagcounter=0 ;

 int tagdict=0 ; 

 int wordcounter=0 ;

 unsigned
   char attributename=0 ;

 unsigned
   char attributevalue=0 ;

 unsigned 
   char typesearchattribute=XCDE_EXACT ;

 unsigned 
   char typesearchvalue=XCDE_EXACT ;

 unsigned
   char typecaseattribute=XCDE_CASE ;

 unsigned
   char typecasevalue=XCDE_CASE ;

 unsigned 
   char typesearch=XCDE_EXACT ;

 unsigned
   char typecase=XCDE_NOCASE ;

 int *tokendict=NULL ;

 char *attributetosearch=NULL ;

 char *valuetosearch=NULL ;

 char *valueattributetosearch=NULL ;

 int numtoken=0 ;
 
 int IL_num=0 ;

 int rank=0 ; 

 int Total_attr_IL_num=0 ;

 int Total_tag_num=0 ;

 int Total_word_num=0 ;

 int New_total_tag_num=0 ;
 
 int New_total_attr_IL_num=0 ;

 int New_total_word_num=0 ;

 int New_total_num=0 ;

 int IL_Attr_num=0 ; 

 int IL_Attrs_num=0 ;

 int New_IL_Attrs_num=0 ; 

 int Selected_Tag_Attr_IL_num=0 ; 

 int IL_Tag_num=0 ;

 int IL_Tags_num=0 ;

 int IL_Word_num=0 ; 

 int New_IL_Word_num=0 ; 

 int New_IL_Tags_num=0 ; 

 int IL_Words_num=0 ; 

 int plane_sweep_proximity_windows_results_num=0 ; 

 int New_plane_sweep_proximity_windows_results_num=0 ; 

 XCDE_IL_Couple_Type *IL_Attr=NULL ; 

 XCDE_IL_Couple_Type *IL_Attrs=NULL ; 

 XCDE_IL_Couple_Type *New_IL_Attrs=NULL ; 

 XCDE_IL_Couple_Type *Selected_Tag_Attr_IL=NULL ;
 
 XCDE_IL_Couple_Type *IL_Tag=NULL ;

 XCDE_IL_Couple_Type *IL_Tags=NULL ; 

 XCDE_IL_Couple_Type *New_IL_Tags=NULL ; 

 XCDE_IL_Triple_Type *IL_Word=NULL ; 

 XCDE_IL_Triple_Type *New_IL_Word=NULL ; 

 XCDE_IL_Triple_Type *IL_Words=NULL ; 

 XCDE_IL_Triple_Type **All_words_lists=NULL ;

 XCDE_IL_Couple_Type *plane_sweep_proximity_windows_results=NULL ; 

 XCDE_IL_Couple_Type *New_plane_sweep_proximity_windows_results=NULL ; 

 int *All_words_lists_lengths=NULL ; 
 
 int All_words_dimension=0 ; 

 int numerror=0 ;

 unsigned 
   char attrwith=0 ;

 unsigned
   char attrwithout=0 ;

 unsigned 
   char attribute=0 ;

 unsigned 
   char tagdist=0 ;

 int attributeerror=0 ;

 int valueerror=0 ;

 int proximity_num=0 ;

 int numdist=0 ;

 int ranktag[2] ;

 int *enclosing=NULL ;

 int number_result=0 ;

 int num_enclosing=0 ;

 XCDE_Console_Type XCDE_Console ; 

 if (argc < 3) { 

   fprintf(stderr, "\t------------------------------------------------------------------") ; 
   
   fprintf(stderr, "\n\t\t       XCDE_SEARCH : Version 1.0BETA 11-07-2002") ;
   
   fprintf(stderr, "\n\t\t         X.ml C.ompressed D.ocuments E.ngine") ; 
   
   fprintf(stderr, "\n\t\t         Paolo Ferragina - Andrea Mastroianni ") ;
   
   fprintf(stderr, "\n\t\t    Dipartimento di Informatica - Universita' di Pisa ") ;
 
   fprintf(stderr, "\n\t\t                 XCDE Search Program") ;
 
   fprintf(stderr, "\n\t------------------------------------------------------------------\n\n") ; 
  
   if (argc != 1) 

     { fprintf(stderr, XCDE_ERROR_PARAMETERS) ;
   
     exit(-1) ; 
     }
   
   if (argc == 1) 

     exit(-1) ;
 }

 // opening console
 res=XCDE_Init_Console(argv[1],
		       &XCDE_Console) ; 
 
 if (res == -1) { 

   fprintf(stderr, "\nFatal Error: problems with the XCDE_Init_Console function\n") ;
   
   exit(-1) ; 
 }
 
 // checks the presence of tagstruct
 res=XCDE_SettedTagstruct(&tagstruct_flag,
			  &XCDE_Console) ; 

 if (res == -1) { 

   fprintf(stderr, "\nFatal Error: problems with the XCDE_Setted_Tagstruct function\n") ;
   
   XCDE_Finalize_Console(&XCDE_Console) ; 
   
   exit(-1) ; 
 }
 
 // if the tagstruct is not present prints a message error and exits
 if (!tagstruct_flag) { 

   fprintf(stderr, "\nFatal Error: the tagstruct file is not present\n") ; 

   XCDE_Finalize_Console(&XCDE_Console) ;

   exit(-1) ;
 }

 // checks the presence of proximity lists
 res=XCDE_SettedProximity(&proximity_flag,
			  &XCDE_Console) ; 
 if (res == -1) { 

   fprintf(stderr, "\nFatal Error: problems with the XCDE_Setted_Proximity function\n") ;
   
   XCDE_Finalize_Console(&XCDE_Console) ; 
   
   exit(-1) ; 
 }	 

 // if the proximity list is not present prints a message error and exit
 if (!proximity_flag) { 

   fprintf(stderr, "\nFatal Error: the proximity lists aren't present\n") ; 
   
   XCDE_Finalize_Console(&XCDE_Console) ;
   
   exit(-1) ;
 }

 numarg=2 ; 
 
 // analizes the command line to execute the query 
 while((argc > numarg) && ((strcmp(argv[numarg], XCDE_TAG_OPTION) == 0) ||
			   (strcmp(argv[numarg], XCDE_TAG_KEYWORD) == 0))) { // if enters this block the command is a tag specification --tag o -t

   // checks if there's a tag specification 
   if (argc > numarg+1) { // if enters this block there's a tag specification 
     
     // increases the pointer to the element analyzed 
     ++numarg ;
     
     // increases the number of tags
     ++tagcounter ; 
     
     // initializes the variable used for tag searches 
     tagdist=0 ; 
     
     numdist=0 ; 
     
     attribute=0 ;
     
     numerror=0 ;
   }
   
   else { // if enters this block there isn't a tag specification, so prints an error message and exits 
     
     fprintf(stderr, "%s.Tag option number %d is not valid\n", XCDE_ERROR_PARAMETERS, tagcounter) ;
     
     fflush(stderr) ;
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
     exit(-1) ; 
   }

   // checks the type of the search for the tag pattern. It may be:
   // --xml_exact o -e 

   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_EXACT_KEYWORD)==0) || (strcmp(argv[numarg], XCDE_EXACT_OPTION) == 0))) { // if enters this block the
                                                                                                                               // search for the tag pattern is exact

     // increases the pointer to the element analyzed
     ++numarg ;
       
     // checks the pattern for the tag is present after the keyword --xml_exact o -e
     if (argc == numarg) { // if enters this block the pattern for the tag isn't present. It prints a 
                           // message error and exits 

       fprintf(stderr, "%s.Tag option number %d is not valid\n", XCDE_ERROR_PARAMETERS, tagcounter) ;
       
       fflush(stderr) ;
       
       XCDE_Finalize_Console(&XCDE_Console) ;
       
       exit(-1) ; 
     }
       
     // token to search into the dictionary 
     tokentosearch=argv[numarg] ;
     
     // increases the pointer to the element analyzed
     ++numarg ;
   }

   // searches the exact tag in the dictionary. If tagdict is 0 so there're no occurrencies,
   // else it will contain the number of the element in the tags dictionary 
   res=XCDE_Search_Tag_Dictionary((char *) tokentosearch,
				  &tagdict, 
				  &XCDE_Console) ;
   
   if (res == -1) { 
     
     fprintf(stderr, "\nFatal Error: problems with the XCDE_Search_Tag_Dictionary function\n") ;
     
     fflush(stderr) ;
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
     exit(-1) ; 
   }
   
   // checks if there's any occurrence 
   if (tagdict == 0) { // if enters this block there is not an occurrence, so destroys
                       // the console and exit

     // frees the memory allocated
     if (IL_Tags != NULL)
       
       free(IL_Tags) ; 
     
     // destroys the console 
     XCDE_Finalize_Console(&XCDE_Console) ;
       
     return(1) ; 
   }
   
   // transforms the dictionary number element into the general rank. Memorizes the rank of the 
   // tag in the ranktag structures to solve distance based tags queries
   res=XCDE_Dict2Rank(TAG,
		      tagdict,
		      &ranktag[((tagcounter-1)%2)],
		      &XCDE_Console) ;
   
   if (res == -1) { 

     fprintf(stderr, "\nFatal Error: problems with the XCDE_Dict2Rank function\n") ;
     
     fflush(stderr) ;
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
     exit(-1) ; 
   }
       
   // gets the inverted list of the tag 
   res=XCDE_Get_IL_ListPos(ranktag[((tagcounter-1)%2)], 
			   &IL_Tag,
			   &IL_Tag_num,
			   &XCDE_Console) ;
   
   // checks if there's any occurrence of the tag 
   if (tagdict == 0)
     
     { // if enters this block there're no occurencies of the tag  

       if (IL_Tags != NULL)
	 
	 free(IL_Tags) ; 
       
       XCDE_Finalize_Console(&XCDE_Console) ;
       
       return(1) ; 
     }

   if (res == -1) { 

     fprintf(stderr, "\nFatal Error: problems with the XCDE_Get_IL_ListPos function\n") ;
     
     fflush(stderr) ;
     
     XCDE_Finalize_Console(&XCDE_Console) ;
       
     exit(-1) ; 
   }
     
   // checks if there's a query on attributes associated with the tag. It can be about the presence of ore or more attributes (--withattr o -a) or 
   // about the absence of ore or more attributes (--withoutattr or -q)
   while ((argc > numarg) && ((strcmp(argv[numarg], XCDE_WITH_ATTRIBUTE_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_WITHOUT_ATTRIBUTE_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_WITH_ATTRIBUTE_OPTION) == 0) || (strcmp(argv[numarg], XCDE_WITHOUT_ATTRIBUTE_OPTION) == 0))) { // if enters this block there's, associated with the tag query, a query about the presence or absence of one or more attributes 

     // ..
     // ..
     // ..  TAGS - ATTRIBUTES SPECIFICATION QUERY 
     // ..
     // ..

     // checks if it's a query about the PRESENCE of an attribute (--withattr or -a)
     if ((strcmp(argv[numarg], XCDE_WITH_ATTRIBUTE_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_WITH_ATTRIBUTE_OPTION) == 0)) { 
     
       // sets the flag for the query about
       // the presence of an attribute 
       attrwith=1;
       
       attrwithout=0 ;
       
       numerror=0 ;
     }
       
     else { // if enters this block it's a query about the absence (--withoutattr or -q) of an attribute

       // sets the flag for the query
       // about the presence of an attribute
       attrwith=0 ;
       
       attrwithout=1 ;
     }
       
     // increases the pointer to the element analyzed
     ++numarg ;
     
     // initializes variables needed 
     attribute=1 ;
     
     attributename=0 ;
       
     attributevalue=0 ;
     
     attributeerror=0 ;

     valueerror=0 ;
     
     valuetosearch=NULL ;
     
     attributetosearch=NULL ;
     
     typesearchattribute=XCDE_ANYVALUE ;
       
     typesearchvalue=XCDE_ANYVALUE ;
     
     // checks if the pattern for the attribute has been specified 
     if (argc == numarg) 
       
       ATTRIBUTE_ERROR ;
     
     // checks the type of search pattern for the attribute. It can be:
     // --xml_anyvalue or -u : for all possible values 
     // --xml_prefix or -l: to specify only the prefix of the attribute name 
     // --xml_suffix or -s: to specify only the suffix of the attribute name 
     // --xml_exact or -e: to specify exactly an attribute name 
     // --xml_contained or -c: to specify a substring that the attribute name must contain 
     // --xml_regexp or -r: to specify an attribute name by a regular expression 
     // --xml_error or -i: to specify an attribute name with an error 

     // checks if it's an anyvalue search pattern
     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ANYVALUE_OPTION)==0) || (strcmp(argv[numarg], XCDE_ANYVALUE_KEYWORD)==0))) { // if enters this block
                                                                                                                                     // it's an anyvalue search pattern
       // sets the type of search to XCDE_ANYVALUE
       typesearchattribute=XCDE_ANYVALUE ;
       
       // increases the pointer to the element analyzed
       ++numarg ;
       
       // sets to NULL the attribute name specified flag
       attributename=0 ;
     }
     
     else
	 
       // checks if it's an attribute prefix specification (--xml_prefix or -l)
       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_PREFIX_OPTION)==0) || (strcmp(argv[numarg], XCDE_PREFIX_KEYWORD)==0))) { // if enters this block
	                                                                                                                           // it's a prefix attribute name specification 

	 // sets to XCDE_PREFIX the type of search 
	 typesearchattribute=XCDE_PREFIX ;
	 
	 // increases the pointer to the element analyzed 
	 ++numarg ;
	 
	 // checks if the prefix of the attribute name is specified
	 if (argc == numarg) 
	   
	   ATTRIBUTE_ERROR ;
	 
	 // points to the attribute prefix 
	 attributetosearch=argv[numarg] ;
	 
	 // sets to 1 the attribute name specified flag 
	 attributename=1 ;
	 
	 // increases the pointer to the element analyzed
	 ++numarg ;
       }
       
       else
	 
	 // checks if it's an attribute name suffix specification (--xml_suffix or -c)
	 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_SUFFIX_OPTION) == 0) || (strcmp(argv[numarg], XCDE_SUFFIX_KEYWORD) == 0))) { // it enters this block it's a
	                                                                                                                                 // suffix name specification 

	   // sets to XCDE_SUFFIX the type of search 
	   typesearchattribute=XCDE_SUFFIX ;
	     
	   // increases the pointer to the element analyzed
	   ++numarg ;
	     
	   // checks if the suffix of the attribute name is specified
	   if (argc == numarg) 
	       
	     ATTRIBUTE_ERROR ;	     
	       
	   // points to the attribute suffix  
	   attributetosearch=argv[numarg] ;
	   
	   // sets to on the attribute name specified flag 
	   attributename=1 ;
	   
	   // increases the pointer to the element analyzed
	   ++numarg ;
	 }
     
	 else
	   
	   // checks if it's an attribute name specified exactly (--xml_exact or -e)
	   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_EXACT_KEYWORD)==0) || (strcmp(argv[numarg], XCDE_EXACT_OPTION) == 0))) { // if enters this block it's an
	                                                                                                                               // attribute name specified exactly

	     // sets to XCDE_EXACT the type of search 
	     typesearchattribute=XCDE_EXACT ;
	     
	     // increases the pointer to the element analyzed
	     ++numarg ;
	   
	     // checks if the attribute name is specified 
	     if (argc == numarg) 
	       
	       ATTRIBUTE_ERROR ;
	     
	     // points to the attribute name 
	     attributetosearch=argv[numarg] ;
	     
	     // sets to on the attribute name specified flag 
	     attributename=1 ;
	     
	     // increases the pointer to the element analyzed
	     ++numarg ;
	   }
     
	   else
	     
	     // checks if it's a substring of the attribute name specification 
	     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CONTAINED_OPTION)==0) ||(strcmp(argv[numarg], XCDE_CONTAINED_KEYWORD)==0))) { // it's an attribute name
	                                                                                                                                      // substring specification 

	       // sets the type of the search  
	       typesearchattribute=XCDE_CONTAINED ;
		 
	       // increases the pointer to the element analyzed 
	       ++numarg ;
		 
	       // checks if the substring of the name is specified 
	       if (argc == numarg) 
		
		 ATTRIBUTE_ERROR ;

	       // points to the substring
	       attributetosearch=argv[numarg] ;
		     
	       // sets to on the attribute name specified flag 
	       attributename=1 ;
		
	       // increases the pointer to the element analyzed
	       ++numarg ;
	     }
       
	     else
		 
	       // checks if it's an attribute name specified by a regular expression (--xml_regexp or -r)
	       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_REGEXP_OPTION)==0) || (strcmp(argv[numarg], XCDE_REGEXP_KEYWORD)==0))) { // if enters this block the
		                                                                                                                           // attribute name is specified 
		                                                                                                                           // by a regular expression 
		 // sets the type of the search 
		 typesearchattribute=XCDE_REGEXP ;
		   
		 // increases the pointer to the element analyzed
		 ++numarg ;
		 
		 // checks if the regular expression is specified
		 if (argc == numarg) 
		   
		   ATTRIBUTE_ERROR ;
		   
		 // points to the regexp 
		 attributetosearch=argv[numarg] ;
		   
		 // sets to on the attribute name specified flag
		 attributename=1 ;
		   
		 // increases the pointer to the element analyzed
		 ++numarg ;
	       }

	       else

		 // checks if it's an attribute name specified by an error (--xml_error or -i)
		 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ERROR_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_ERROR_OPTION) == 0)))
		   
		   { // increases the pointer to the element analyzed 
		     ++numarg ;
		     
		     // checks if the number of error is specified 
		     if (argc == numarg) 
		     
		       ATTRIBUTE_ERROR ;
		   
		     // gets the number of error on the string 
		     attributeerror=atoi(argv[numarg]) ;
		   
		     // checks if the number of error is specified exactly 
		     if (attributeerror <= 0) { // if enters this block the number of error isn't specified exactly 

		       fprintf(stderr, "%s.Attribute option of tag number %d is not valid\n", XCDE_ERROR_PARAMETERS, tagcounter) ;
		     
		       fflush(stderr) ;
		       
		       if (IL_Tag != NULL)
			 
			 free(IL_Tag) ; 
		     
		       fflush(stderr) ;
		       
		       XCDE_Finalize_Console(&XCDE_Console) ;
		       
		       exit(-1) ; 
		     }
		     
		     // increases the pointer to the element analyzed 
		     ++numarg ;
		     
		     // checks if the attribute name is specified
		     if (argc == numarg) 
		       
		       ATTRIBUTE_ERROR ;
		     
		     // sets the type of search 
		     typesearchattribute=XCDE_ERROR ;

		     // points to the attribute name 
		     attributetosearch=argv[numarg] ;
		     
		     // sets to on the attribute name specified flag 
		     attributename=1 ;
		     
		     // increases the pointer to the element analyzed
		     ++numarg ;
		   }

		 else // in this case it's a correct type of search, so prints an error message and exits 

		   ATTRIBUTE_ERROR ;
       
     // if the type of search is different from XCDE_ANYVALUE looks for the type of case sensitivity in 
     // the search (it's optional and can be --xml_case or -c for the case sensitive and --xml_nocase or -n
     // for case insensitive). In the case it is't specified the search is for default case insensitive

     if (typesearchattribute != XCDE_ANYVALUE) { // if enters this block the type of search is different from XCDE_ANYVALUE and so 
                                                 // looks if there's the case sensitivity option 
       
       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CASE_OPTION)==0) || (strcmp(argv[numarg], XCDE_CASE_KEYWORD)==0))) { // if enters this block it's 
	                                                                                                                       // present the XCDE_CASE option 
	 // sets the type of search as XCDE_CASE
	 typecaseattribute=XCDE_CASE ;
	 
	 // increases the pointer to the element analyzed
	 ++numarg ;
       }
       
       else
	 
	 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_NOCASE_OPTION)==0) || (strcmp(argv[numarg], XCDE_NOCASE_KEYWORD)==0))) { 

	   // sets the type of search as XCDE_NOCASE
	   typecaseattribute=XCDE_NOCASE ;
	   
	   // increases the pointer to the element analyzed
	   ++numarg ; 
	 }
     }

     // checks if the value of attributes has been specified (--value or -v) 
     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ATTRIBUTE_VALUE_OPTION) == 0) || (strcmp(argv[numarg], XCDE_ATTRIBUTE_VALUE_KEYWORD) == 0))) { // if enters this block the value of 
                                                                                                                                                       // the attributes has been specified 

       // increases the pointer to the element analyzed
       ++numarg ;
       
       // checks if the value of the attributes is specified 
       if (argc == numarg)
	 
	 ATTRIBUTE_ERROR ; 
       
       // checks if the modificator for anyvalue (that is the standard) has been inserted 
       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ANYVALUE_OPTION)==0) || (strcmp(argv[numarg], XCDE_ANYVALUE_KEYWORD)==0))) { 

	 // sets the type of search for XCDE_ANYVALE
	 typesearchvalue=XCDE_ANYVALUE ;
	 
	 // increases the pointer to the element analyzed 
	 ++numarg ;
	 
	 // sets to off the flag for the attribute value specification 
	 attributevalue=0 ;
       }
       
       else
	 
	 // checks if the value is specified by a prefix pattern search (--xml_prefix or -l)
	 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_PREFIX_OPTION)==0) || (strcmp(argv[numarg], XCDE_PREFIX_KEYWORD)==0))) { // if enters this block the value of the 
	   // attribute is specified by a prefix pattern search 
	   
	   // sets the type of search for the attribute value to prefix pattern search 
	   typesearchvalue=XCDE_PREFIX ;
	   
	   // increases the pointer to the element analyzed 
	   ++numarg ;
	   
	   // checks if the prefix pattern is specified
	   if (argc == numarg) 
	     
	     ATTRIBUTE_ERROR ;
	   
	   // points to the prefix pattern 
	   valuetosearch=argv[numarg] ;
	   
	   // sets to on the flag for the attribute value specification 
	   attributevalue=1 ;

	   // increases the pointer to the element analyzed 
	   ++numarg ;
	 }
       
	 else
	   
	   // checks if the value is specified by a suffix pattern search (--xml_suffix or -s) 
	   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_SUFFIX_OPTION)==0) || (strcmp(argv[numarg], XCDE_SUFFIX_KEYWORD)==0))) { // if enters this block the pattern is specified
	                                                                                                                               // by a suffix pattern search 

	     // sets the type of search for the value of the attribute to suffix
	     typesearchvalue=XCDE_SUFFIX ;
	     
	     // increases the pointer to the element analyzed
	     ++numarg ;
	     
	     // checks if the suffix of the attribute value has been specified 
	     if (argc == numarg) 
	       
	       ATTRIBUTE_ERROR ; 
	     
	     // pointes to the suffix of the attribute value 
	     valuetosearch=argv[numarg] ;
	   
	     // puts to on the attribute value specification flag
	     attributevalue=1 ;
	     
	     // increases the pointer to the element analyzed 
	     ++numarg ;
	   }
     
	   else

	     // checks if the value is specified by an exact pattern search (--xml_exact or -e)
	     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_EXACT_OPTION)==0) || (strcmp(argv[numarg], XCDE_EXACT_KEYWORD) == 0))) { // if enters this block the
	                                                                                                                                 // value is specified by an exact pattern search 

	       // sets the type of search to xcde_exact 
	       typesearchvalue=XCDE_EXACT ;
	     
	       // increases the pointer to the element analyzed
	       ++numarg ;
	     
	       // checks if the value of the attribute has been specified 
	       if (argc == numarg) 
		 
		 ATTRIBUTE_ERROR ;
	     
	       // points to the exact value to search 
	       valuetosearch=argv[numarg] ;
	
	       // sets to on the flag for the attribute value specification 
	       attributevalue=1 ;
	       
	       // increases the pointer to the element analyzed
	       ++numarg ;
	     }
       
	     else
	       
	       // checks if the value is specified by a substring prefix pattern search (--xml_contained or -c)
	       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CONTAINED_OPTION)==0) || (strcmp(argv[numarg], XCDE_CONTAINED_KEYWORD)==0))) { // if enters this block
		                                                                                                                                 // the value of the attribute has been specified with a
		                                                                                                                                 // substring pattern search 
		 
		 // sets the type of search as a substrin g
		 typesearchvalue=XCDE_CONTAINED ;
	       
		 // increases the pointer to the element analyzed
		 ++numarg ;
	       
		 // checks if prefix has been specified 
		 if (argc == numarg) 
		   
		   ATTRIBUTE_ERROR ;
	       
		 // points to the substring 
		 valuetosearch=argv[numarg] ;
	       
		 // sets to on the attribute value specified flag
		 attributevalue=1 ;
	       
		 // increases the pointer to the element analyzed
		 ++numarg ;
	       }
       
	       else
		 
		 // checks if the value is specified by a regular expression prefix pattern search (--xml_regexp or -r)
		 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_REGEXP_OPTION)==0) || (strcmp(argv[numarg], XCDE_REGEXP_KEYWORD)==0))) { // if enters this block
		                                                                                                                             // the value of the attribute has been specified with a 

		   // sets the type of search 
		   typesearchvalue=XCDE_REGEXP ;
		 
		   // increases the pointer to the element analyzed
		   ++numarg ;
		 
		   // checks if the regular expression has been specified 
		   if (argc == numarg) 
		     
		     ATTRIBUTE_ERROR ;
		   
		   // points to the regular expression 
		   valuetosearch=argv[numarg] ;
		 
		   // sets to on the attribute value specified flag 
		   attributevalue=1 ;
		 
		   // increases the pointer to the element analyzed
		   ++numarg ;
		 }
       
		 else
		   
		   // checks if the value is specified by a token exact pattern search (--xml_token_exact or -w)
		   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_TOKEN_EXACT_OPTION)==0) || (strcmp(argv[numarg], XCDE_TOKEN_EXACT_KEYWORD)==0))) { // if enters this block the value is specified
		                                                                                                                                         // by a token exact pattern search 

		     // sets the type of the search to token exact pattern search 
		     typesearchvalue=XCDE_TOKEN_EXACT ;
		   
		     // increases the pointer to the element analyzed
		     ++numarg ;
		   
		     // checks if the regular expression has been specified 
		     if (argc == numarg) 
		       
		       ATTRIBUTE_ERROR ;
		     
		     // points to the token exact value 
		     valuetosearch=argv[numarg] ;
		   
		     // sets to on the attribute value specified flag 
		     attributevalue=1 ;
		     
		     // increases the pointer to the element analyzed
		     ++numarg ;
		   }
       
		   else
		     
		     // checks if the value is specified by a token prefix pattern search (--xml_token_prefix or -y)
		     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_TOKEN_PREFIX_OPTION)==0) || (strcmp(argv[numarg], XCDE_TOKEN_PREFIX_KEYWORD)==0))) { // if enters this block the value is specified by a 
		                                                                                                                                             // token prefix pattern search 

		       // sets the type of search to token prefix pattern search 
		       typesearchvalue=XCDE_TOKEN_PREFIX ;

		       // increases the pointer to the element analyzed
		       ++numarg ;
		       
		       // checks if the regular expression has been specified 
		       if (argc == numarg) 
			 
			 ATTRIBUTE_ERROR ;
		       
		       // points to the token prefix value 
		       valuetosearch=argv[numarg] ;
		       
		       // sets to on the attribute value specified flag 
		       attributevalue=1 ;
		       
		       // increases the pointer to the element analyzed 
		       ++numarg ;
		     }
       
		     else
		 
		       // checks if the value is specified by a token suffix pattern search (--xml_token_suffix or -z)
		       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_TOKEN_SUFFIX_OPTION)==0) || (strcmp(argv[numarg], XCDE_TOKEN_SUFFIX_KEYWORD) == 0))) { 

			 // sets the type of search to token suffix pattern search 
			 typesearchvalue=XCDE_TOKEN_SUFFIX ;
			 
			 // increases the pointer to the element analyzed
			 ++numarg ;
			 
			 // checks if the suffix token has been specified 
			 if (argc == numarg) 
			   
			   ATTRIBUTE_ERROR ;
		       
			 // points to the value to search 
			 valuetosearch=argv[numarg] ;
		       
			 // sets to on the attribute value specified flag
			 attributevalue=1 ;
		     
			 // increases the pointer to the elment analyzed
			 ++numarg ;
		       }
       
		       else
		       
			 // checks if the value is specified by a search pattern with errors 
			 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ERROR_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_ERROR_OPTION) == 0))) { // if enters this block the value is specified by a search 
			                                                                                                                               // pattern with errors 
			   // sets the type of search with errors 
			   typesearchvalue=XCDE_ERROR ; 

			   // increases the pointer to the element analyzed
			   ++numarg ;
			   
			   // checks if the number of errors in the pattern has been specified 
			   if (argc == numarg) 
			     
			     ATTRIBUTE_ERROR ;
				 
			   // reads the number of errors in the pattern 
			   valueerror=atoi(argv[numarg]) ;

			   // checks if the number of errors is correct 
			   if (valueerror <= 0) { // if enters this block the number of errors isn't correct 

			     fprintf(stderr, XCDE_ERROR_PARAMETERS) ; 
			     
			     fflush(stderr) ;
			     
			     XCDE_Finalize_Console(&XCDE_Console) ;
			     
			     exit(-1) ; 
			   }
			   
			   // increases the pointer to the element analyzed
			   ++numarg ;
			 
			   // checks if the pattern with errors has been specified 
			   if (argc == numarg) 
			     
			     ATTRIBUTE_ERROR ;
			   
			   // points to the pattern with errors 
			   valuetosearch=argv[numarg] ;

			   // sets to on the attribute value specified flag
			   attributevalue=1 ;
			 
			   // increases the pointer to the element analyzed
			   ++numarg ;
			 }
       
			 else // in this case the type of search for the attribute value hasn't been specified correctly
			   
			   ATTRIBUTE_ERROR ;
       
       // checks if the type of search for the attribute value is different from XCDE_ANYVALUE 
       if (typesearchvalue != XCDE_ANYVALUE) { // if enters this block the type of search for the attribute value is different from XCDE_ANYVALUE
	                                       // so it may be specified the case sensitivity of the search 

	 // checks if the case sensitivity has been specified (--xml_case or -c)
	 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CASE_OPTION)==0) || (strcmp(argv[numarg], XCDE_CASE_KEYWORD)==0))) { // if enters this block the 
	                                                                                                                         // case sensitivity has been specified 

	   typecasevalue=XCDE_CASE ;
	   
	   // increases the pointer to the element analyzed
	   ++numarg ;
	 }
	 
	 else
	     
	   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_NOCASE_OPTION)==0) || (strcmp(argv[numarg], XCDE_NOCASE_KEYWORD) == 0))) { // if enters this block the nocase
	                                                                                                                                 // sensitivity hasn't been specified (--xml_nocase or -n) 

	     // sets the type of the search
	     typecasevalue=XCDE_NOCASE ;
	     
	     // increases the pointer to the element analyzed 
	     ++numarg ; 
	   }
       }
     }
     
     
     // checks if an attribute name or value has been specified. The name or the value is different from
     // XCDE_ANYVALUE and so it must be found in the attribute dictionary 
     if ((attributename) || (attributevalue)) { // if enters this block an attribute name or value has been specified 
 	 
       // search the attribute name or value in the attribute dictionary and puts the number of attributes
       // found in numtoken and its ranks in the array tokendict 
       res=XCDE_Search_Attr_Dictionary((char *) attributetosearch,
				       (char *) valuetosearch, 
				       &tokendict,
				       &numtoken,
				       typesearchattribute,
				       typesearchvalue,
				       typecaseattribute, 
				       typecasevalue, 
				       attributeerror,
				       valueerror,
				       &XCDE_Console) ;
	 
       if (res == -1) { // checks if there's a problem with the search in the dictionary 

	 fprintf(stderr, "\nFatal Error: problems with the XCDE_Search_Attr_Dictionary function\n") ;
	   
	 fflush(stderr) ;
	 
	 if (IL_Tag != NULL)
	   
	   free(IL_Tag) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
       }
     }
     
     else 
	   
       // the name or the value has been specified, so loads all the attributes elements  
       { // gets the number of elements in the dictionary 
	 res=XCDE_Get_Num_El_Dict(ATTR, 
				  &numtoken, 
				  &XCDE_Console) ;
	 
	 if (res == -1) { 
	   
	   fprintf(stderr, "\nFatal Error: problems with the XCDE_Get_Num_El_Dict function\n") ;
	   
	   fflush(stderr) ;
	   
	   if (IL_Tag != NULL)
	     
	     free(IL_Tag) ; 
	   
	   if (IL_Tags != NULL)
	     
	     free(IL_Tags) ; 
	   
	   XCDE_Finalize_Console(&XCDE_Console) ;
	   
	   exit(-1) ; 
	 }
       }
     
     // loops over all the attributes found 
     for(k=0 ; k<numtoken; ++k) { 

       // checks if an attribute name or value has been specified 
       if ((attributename) || (attributevalue)) 
	 
	 // in this case an element has been found and so founds its rank
	 res=XCDE_Dict2Rank(ATTR,
			    tokendict[k], 
			    &rank,
			    &XCDE_Console) ;
       
       else
	 
	 // its an XCDE_ANYVALUE element. Founds its rank 
	 res=XCDE_Dict2Rank(ATTR,
			    k+1,
			    &rank,
			    &XCDE_Console) ;
       
       
       if (res == -1) { 
	 
	 fprintf(stderr, "\nFatal Error: problems with the XCDE_Dict2Rank function\n") ;
	 
	 fflush(stderr) ;
	     
	 if (IL_Tag != NULL)
	   
	   free(IL_Tag) ;
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	     XCDE_Finalize_Console(&XCDE_Console) ;
	     
	     exit(-1) ; 
       }
       
       // gets the inverted list of the elements
       res=XCDE_Get_IL_ListPos(rank,
			       &IL_Attr,
			       &IL_Attr_num,
			       &XCDE_Console) ;
       
       if (res == -1) { 
	 
	 fprintf(stderr, "\nFatal Error: problems with the XCDE_Get_IL_ListPos function\n") ;
	 
	 fflush(stderr) ;
	 
	 if (IL_Tag != NULL)
	   
	   free(IL_Tag) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (IL_Attr != NULL)
	   
	   free(IL_Attr) ; 
	 
	 if (IL_Attrs != NULL)
	   
	   free(IL_Attrs) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
       }
       
       // join the inverted lists of the attributes, the lists are joint in a new 
       // sorted list 
       res=join_attrs_or_words(&New_IL_Attrs,
			       &New_IL_Attrs_num,
			       IL_Attrs,
			       IL_Attrs_num,
			       IL_Attr,
			       IL_Attr_num) ; 
       
       if (res == -1) { 
	 
	 fprintf(stderr, "\nFatal Error: problems with the join_attrs_or_words function\n") ;
	 
	 fflush(stderr) ;
	 
	 if (IL_Tag != NULL)
	   
	   free(IL_Tag) ;
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (IL_Attr != NULL)
	   
	   free(IL_Attr) ; 
	 
	 if (IL_Attrs != NULL)
	   
	   free(IL_Attrs) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
       }
       
       // frees the memory allocated for the old list 
       if (IL_Attrs != NULL) { 
	 
	 free(IL_Attrs) ; 
	 
	 IL_Attrs=NULL; 
       }
       
       // copies the new list 
       IL_Attrs=New_IL_Attrs ;
	 
       IL_Attrs_num=New_IL_Attrs_num ; 
       
       // frees the memory allocated for the old list 
       if (IL_Attr != NULL) { 
	 
	 free(IL_Attr) ;
	 
	 IL_Attr=NULL ;
       }
       
       // set to 0 or NULL the counter and pointers used to copy lists 
       IL_Attr_num=0 ; 
       
       New_IL_Attrs=NULL ; 
	
       New_IL_Attrs_num=0 ; 
     }
     
     // frees the memory allocated for the list of ranks 
     if (tokendict != NULL) { 

       free(tokendict) ;
       
       tokendict=NULL ;
     }
     
     // sets to 0 the number of tokens found 
     numtoken=0 ; 
       
     // checks if the tag must be with or without the attributes found
     if (attrwith)

       // in this case the tag must be with the attributes founds, so 
       // calls the function to select the tag occurrencies with them 
       res=select_tag_with_attrs(&Selected_Tag_Attr_IL, 
				 &Selected_Tag_Attr_IL_num, 
				 IL_Tag, 
				 IL_Tag_num, 
				 IL_Attrs, 
				 IL_Attrs_num) ; 
     
     else
       
       // selects the elements of TAG IL that don't contain the attribute
       // calls the function to select the tag occurrencies without them 
       res=select_tag_without_attrs(&Selected_Tag_Attr_IL, 
				    &Selected_Tag_Attr_IL_num, 
				    IL_Tag, 
				    IL_Tag_num,
				    IL_Attrs,
				    IL_Attrs_num) ; 
     
     if (res == -1) { 

       fprintf(stderr, "\nFatal Error: problems with the select_tag_without_attrs or select_tag_with_attrs function\n") ;
       
       fflush(stderr) ;
       
       if (IL_Tag != NULL)
	 
	 free(IL_Tag) ; 
       
       if (IL_Tags != NULL)
	 
	 free(IL_Tags) ; 
       
       if (IL_Attrs != NULL)
	 
	 free(IL_Attrs) ; 
       
       if (Selected_Tag_Attr_IL != NULL)
	 
	 free(Selected_Tag_Attr_IL) ; 
       
       XCDE_Finalize_Console(&XCDE_Console) ;
       
       exit(-1) ; 
     }
     
     // frees the memory needed to contain the old list 
     if (IL_Attrs != NULL) { 

       free(IL_Attrs) ;
	   
       IL_Attrs=NULL ;
     }
     
     IL_Attrs_num=0 ; 
     
     // frees the memory needed to contain the old list 
     if (IL_Tag != NULL) { 

       free(IL_Tag) ;
       
       IL_Tag=NULL ;
     }
    
     // copies the new list (with or without the attributes). IL_TAG contains now the position of the tags until the present  
     IL_Tag=Selected_Tag_Attr_IL ; 
     
     IL_Tag_num=Selected_Tag_Attr_IL_num ; 
     
     Selected_Tag_Attr_IL_num=0 ; 
     
     Selected_Tag_Attr_IL=NULL ; 
   }
   
   // It's now possible to query about the distance between tags (--xml_dist or -d)

   // query about the distance beetween the tags. Itsn't possible with the first tag 
   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_DISTANCE_TAG_KEYWORD)==0) || (strcmp(argv[numarg], XCDE_DISTANCE_TAG_OPTION)==0))) { // if enters this block
                                                                                                                                           // it's a query about the distance

     // increases the pointer to the element analyzed
     ++numarg ;
     
     tagdist=1 ;
     
     numdist=0 ;
       
     // checks if it's the first tag. It's not possible to query the distance for the first tag 
     if ((argc == numarg) || (tagcounter == 1)) { // it enters this block is the first tag and isn't possible to query it 

       fprintf(stderr, "%s.It's not possible to insert the distance for the first tag\n", XCDE_ERROR_PARAMETERS) ;
	 
	 fflush(stderr) ;
	 
	 if (IL_Tags != NULL)

	   free(IL_Tags) ; 

	 if (IL_Tag != NULL)

	   free(IL_Tag) ; 

	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
     }
       
     // reads the distance inserted
     numdist=atoi(argv[numarg]) ;
       
     // checks if the distance inserted is correct (it must be greater that 0)
     if (numdist <= 0)  { 

       fprintf(stderr, "%s.Distance number isn't correct at tag number\n", XCDE_ERROR_PARAMETERS, tagcounter) ; 
       
	 fflush(stderr) ;
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (IL_Tag != NULL)
	   
	   free(IL_Tag) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
     }
     
     // increases the pointer to the element analyzed
     ++numarg ;
   }
   
   // checks if the distance has been requested
   if (tagdist) { // if enters this block the distance has been requested

     // selects the tag elements in the second list that have the correct distance. ranktag contains the tag of the previous element  
     res=select_tag_nested_in_tag_with_distance(&New_IL_Tags,
						&New_IL_Tags_num, 
						ranktag[((tagcounter-2)%2)],
						numdist,
						IL_Tags, 
						IL_Tags_num, 
						IL_Tag,
						IL_Tag_num,
						&XCDE_Console) ;
     
     if (res == -1) { 

       fprintf(stderr, "\nFatal Error: problems with the select_nested_in_tag_with_dist function\n") ;
       
       fflush(stderr) ;
       
       if (IL_Tags != NULL)
	 
	 free(IL_Tags) ; 
       
       if (IL_Tag != NULL)
	 
	 free(IL_Tag) ; 
       
       XCDE_Finalize_Console(&XCDE_Console) ;
       
       exit(-1) ; 
     }
     
     // frees the memory needed to contain the first list
     if (IL_Tag != NULL) { 

       free(IL_Tag) ; 
       
       IL_Tag=NULL ; 
     }
     
     IL_Tag_num=0 ; 
     
     // frees the memory needed to contain the second list 
     if (IL_Tags != NULL) { 

       free(IL_Tags) ; 
       
       IL_Tags=NULL ;
     }

     IL_Tags_num=0 ; 
     
     IL_Tags=New_IL_Tags ; 
     
     IL_Tags_num=New_IL_Tags_num ; 
     
     New_IL_Tags_num=0 ;
     
     New_IL_Tags=NULL ; 
   }

   else

     if (tagcounter > 1) { // if enters this block more tags have been specified in the query. Selects 
                           // the occurrencies of the last tag that are contained in the previous. 

       // selects the elements in the second list that are contained in the first list 
       res=select_tag_nested_in_tag(&New_IL_Tags,
				    &New_IL_Tags_num, 
				    IL_Tags,
				    IL_Tags_num, 
				    IL_Tag, 
				    IL_Tag_num) ; 
       
       if (res == -1) { 

	 fprintf(stderr, "\nFatal Error: problems with the select_tag_nested_in_tag function\n") ;
	 
	 fflush(stderr) ;
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
       }
	
       // frees the memory needed to contain the old list
       if (IL_Tags != NULL) { 

	 free(IL_Tags) ; 
	 
	 IL_Tags=NULL ;
       }

       // copies the new list 
       IL_Tags_num=New_IL_Tags_num ;
	   
       IL_Tags=New_IL_Tags ; 

       // frees the memory needed 
       if (IL_Tag != NULL) { 

	 free(IL_Tag) ; 
	 
	 IL_Tag = NULL ;
       }
       
       IL_Tag_num=0 ;
     }
   
   // in this case there's only ONE TAG specified. So copies its IL 
     else 
	 
       {  IL_Tags_num=IL_Tag_num ;

	 IL_Tags=IL_Tag ; 

	 IL_Tag_num=0 ;

	 IL_Tag=NULL ;
       }
 }

 // checks if there's a tag option. All queries must contain at least one tag option 
 /*  if (tagcounter == 0) */

/*     { fprintf(stderr, "\n%sThere must be one or more tag in the query expression\n", XCDE_ERROR_PARAMETERS) ; */

/*     fflush(stderr) ;  */

/*     XCDE_Finalize_Console(&XCDE_Console) ;  */

/*     exit(-1) ;  */
/*     } */

 // ..
 // ..
 // ..  WORDS - PROXIMITIES SPECIFICATION QUERY 
 // ..
 // ..
 
 // IL_Tags contains the position of tags selected and IL_Tags_num contains its number
 
 // ..
 // ..  PROXIMITY OPTION 
 // .. 

 // checks if there's the proximity option (--proximity or -p)
 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_PROX_OPTION) == 0) || (strcmp(argv[numarg], XCDE_PROX_KEYWORD) == 0))) { // if enters this block the proximity
                                                                                                                             // option is present

   // increases the pointer to the analyzed element 
   ++numarg ;
   
   // checks if the number of the proximity is present 
   if (argc == numarg) { // in this case the proximity number isn't present. Prints a message error and exits 
     
     fprintf(stderr, "%s.Proximity option isn't correct\n", XCDE_ERROR_PARAMETERS) ;
     
     fflush(stderr) ;
     
     if (IL_Tags != NULL)
       
       free(IL_Tags) ; 
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
     exit(-1) ; 
   }
   
   // reads the number of the proximity 
   proximity_num = atoi(argv[numarg]) ;
   
   // increases the pointer to the analyzed element 
   ++numarg ;
   
   // checks if the proximity number is correct 
   if (proximity_num <= 0) { // if enters this block the proximity number isn't correct. Prints a message error and exits 

     fprintf(stderr, "%s.Proximity option isn't correct. The proximity must be greater than 0\n", XCDE_ERROR_PARAMETERS) ;
     
     fflush(stderr) ;
     
     if (IL_Tags != NULL)
       
       free(IL_Tags) ; 
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
     exit(-1) ; 
   }
 }

 // ..
 // ..
 // .. WORDS ANALYZE
 // ..
 // ..

 // loops over all the words specifications 
 while((argc > numarg)) { 

   // checks if it's a word specifications (--xml_word or -w)
   if (((strcmp(argv[numarg], XCDE_WORD_OPTION) == 0) ||
	(strcmp(argv[numarg], XCDE_WORD_KEYWORD) == 0)) &&
       (argc > numarg+1)) { // if enters this block it's a word specification 

     // increases the pointer to the analyzed element 
     ++numarg ;
       
     // increases the number of words specifications inserted 
     ++wordcounter ; 
       
     // sets to 0 the number of errors
     numerror=0 ;
       
     // sets to XCDE_NOCASE the case sensitivity of the query 
     typecase=XCDE_NOCASE ;
   }
   
   else // in this case it's not a word specification and it isn't correct
       
     WORD_ERROR ;

   // checks the type of search pattern for the attribute. It can be:
   // --xml_exact or -e: to specify exactly an attribute name 
   // --xml_prefix or -l: to specify only the prefix of the attribute name 
   // --xml_suffix or -s: to specify only the suffix of the attribute name 
   // --xml_contained or -c: to specify a substring that the attribute name must contain 
   // --xml_regexp or -r: to specify an attribute name by a regular expression 
   // --xml_error or -i: to specify an attribute name with an error 
   
   // it's a exact search (--xml_exact or -e)
   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_EXACT_KEYWORD)==0) || (strcmp(argv[numarg], XCDE_EXACT_OPTION) == 0))) { // if enters this block it's an exact search

     // sets to XCDE_EXACT the type of the search
     typesearch=XCDE_EXACT ;
       
     // increases the pointer to the element analyzed
     ++numarg ;
       
     // checks if the words search pattern has been specified, else prints an error message an exits
     if (argc == numarg) 
	 
       WORD_ERROR ; 
       
     // copies the word search patten 
     tokentosearch=argv[numarg] ;
       
     // increases the pointer to the element analyzed
     ++numarg ;
   }
     
     else
       
       // checks if it's a prefix pattern seach (--xml_prefix or -l)
       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_PREFIX_KEYWORD)==0) || (strcmp(argv[numarg], XCDE_PREFIX_OPTION) == 0))) { // if enters this block it's a prefix search

	 // sets the type of search to XCDE_PREFIX
	 typesearch=XCDE_PREFIX ;
	 
	 // increases the pointer to the element analyzed
	 ++numarg ;
	 
	 // checks if the words search pattern has been specified, else prints an error message an exits
	 if (argc == numarg) 
	   
	   WORD_ERROR ; 
	 
	 // copies the search pattern  
	 tokentosearch=argv[numarg] ;
	 
	 // increases the pointer to the element analyzed
	 ++numarg ;
       }
     
       else
	 
	 // checks if it's a suffix search (--xml_suffix or -s)
	 if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_SUFFIX_KEYWORD)==0 ) || (strcmp(argv[numarg], XCDE_SUFFIX_OPTION) == 0))) { // if enters this block it's a 
	                                                                                                                                // suffix search

	   // sets the type of the search to XCDE_SUFFIX 
	   typesearch=XCDE_SUFFIX ;
	   
	   // increases the pointer to the element analyzed
	   ++numarg ;
	       
	   // checks if there's the search pattern for the suffix
	   if (argc == numarg) 
	     
	     WORD_ERROR ; 
	   
	   // token to search into the dictionary 
	   tokentosearch=argv[numarg] ;
	   
	   // increases the pointer to the element analyzed
	   ++numarg ;
	   }
   
	 else
	   
	   // checks if it's a search for the substring (--xml_contained or -c)
	   if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CONTAINED_OPTION)==0) || (strcmp(argv[numarg], XCDE_CONTAINED_KEYWORD)==0))) { // if enters this block
	                                                                                                                                     // it's a substring search 
	     
	     // sets the type of search 
	     typesearch=XCDE_CONTAINED ;
	     
	     // increases the pointer to the element analyzed
	     ++numarg ;

	     // checks if there's the search pattern for the string contained, else prints a message error and exits
	     if (argc == numarg) 
	       
	       WORD_ERROR ; 
	     
	     // points to the search pattern for the string contained 
	     tokentosearch=argv[numarg] ;
	     
	     // increases the pointer to the element analyzed
	     ++numarg ;
	     }
     
	   else
	     
	     // checks if it's a search with a search pattern that contains errors (--xml_error or -i)
	     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_ERROR_KEYWORD) == 0) || (strcmp(argv[numarg], XCDE_ERROR_OPTION) == 0))) { // if enters this block
	                                                                                                                                   // it's a search with errors

	       // increases the pointer to the element analyzed
	       ++numarg ;
	       
	       // checks if there's the number of errors, else prints a message error and exits 
	       if (argc == numarg) 
		 
		 WORD_ERROR ; 
	       
	       // reads the number of errors inserted 
	       numerror=atoi(argv[numarg]) ;
	       
	       // checks if the number of errors is correct, else prints a message error and exits 
	       if (numerror <= 0)
		 
		 WORD_ERROR ; 
	       
	       // increases the pointer to the element ananlyzed
	       ++numarg ;

	       // checks if there's the search pattern for the search with errors, else prints a message error and exits	       
	       if (argc == numarg) 
		 
		 WORD_ERROR ; 
	       
	       // sets the type of search 
	       typesearch=XCDE_ERROR ;
	       
	       // points to the pattern to search with the error
	       tokentosearch=argv[numarg] ;
	       
	       // increases the pointer to the element analyzed
	       ++numarg ;
	     }
     
	     else
		 
	       // checks if it's a regular expression search (--xml_regexp or -r) 
	       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_REGEXP_OPTION)==0) || (strcmp(argv[numarg], XCDE_REGEXP_KEYWORD)==0))) { // if enters this block it's 
		                                                                                                                           // a regular expression search 
		 
		 // sets the type of the search 
		 typesearch=XCDE_REGEXP ;
		 
		 // increases the pointer to the element analyzed 
		 ++numarg ;
		 
		 // checks if there's the search pattern for the search with regular expression, else prints a message error and exits 
		 if (argc == numarg) 
		     
		   WORD_ERROR ; 
		   
		 // points to the regular expression to search 
		 tokentosearch=argv[numarg] ;
		 
		 // increases the pointer to the element analyzed
		 ++numarg ;
	       }
   
	       else
		 
		 // it isn't a correct option. Prints an error message and exits
		 WORD_ERROR ; 
     
   // checks if the type of search is different from XCDE_ERROR and XCDE_REGEXP
   if ((typesearch != XCDE_ERROR) && (typesearch != XCDE_REGEXP)) { // if enters this block the type of the search is different from XCDE_ERROR and XCDE_REGEXP,
                                                                    // so it must checks if the case sensitivity for the search has been specified 

     // checks if the case sensitivity for the search has been specified (--xml_case or -c)
     if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_CASE_OPTION) == 0) || (strcmp(argv[numarg], XCDE_CASE_KEYWORD) == 0))) { // if enters this block the
                                                                                                                                 // case sensitivity has been specified

       // sets the type of the search to case sensitive 
       typecase=XCDE_CASE ;
       
       // increases the pointer to the element analyzed
       ++numarg ;
     }
	   
     else
	
       // checks if the case insensitivity for the search has been specified (--xml_nocase or -c)
       if ((argc > numarg) && ((strcmp(argv[numarg], XCDE_NOCASE_OPTION) == 0) || (strcmp(argv[numarg], XCDE_NOCASE_KEYWORD) == 0))) { // if enters this block
	                                                                                                                               // the case insensitivity hasn't been specified 

	 // sets the type of search to case insensitivity 
	 typecase=XCDE_NOCASE ;
	 
	 // increases the pointer to the element analyzed 
	 ++numarg ; 
       }
   }
     
   // search in the words dictionary. Specifies the type of search and its case and numtoken will contain
   // the number of occurrencies and tokendict is an array that will contain the list of occurrencies 
   res=XCDE_Search_Word_Dictionary((char *) tokentosearch,
				   &tokendict,
				   &numtoken,
				   typesearch,
				   typecase, 
				   numerror, 
				   &XCDE_Console) ;
     
   if (res == -1) { 

     fprintf(stderr, "\nFatal Error: problems with the XCDE_Search_Word_Dictionary function\n") ;
     
     fflush(stderr) ;
     
     if (IL_Tags != NULL)
       
       free(IL_Tags) ; 
       
     if (All_words_lists_lengths != NULL) 
	 
       free(All_words_lists_lengths) ; 
     
     for(k=0 ; k<wordcounter; ++k) 
       
       free(All_words_lists[k]) ; 
     
     if (All_words_lists != NULL)
       
       free(All_words_lists) ; 
     
     XCDE_Finalize_Console(&XCDE_Console) ;
     
       exit(-1) ; 
   }
   
   
   for(k=0 ; k<numtoken; ++k) { // loops over all occurrencies 

     // traslates the occurrerce number in the words dictionary to the its rank in the 
     // global dictionary 
     res=XCDE_Dict2Rank(WORD,
			tokendict[k], 
			&rank,
			&XCDE_Console) ;
     
       if (res == -1) { 

	 fprintf(stderr, "\nFatal Error: problems with the XCDE_Dict2Rank function\n") ;
	 
	 fflush(stderr) ;
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
	 
	 if (tokendict != NULL)
	   
	   free(tokendict) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (All_words_lists_lengths != NULL) 
	   
	   free(All_words_lists_lengths) ; 
	 
	 for(k=0 ; k<wordcounter; ++k) 
	   
	   free(All_words_lists[k]) ; 
	 
	 if (All_words_lists != NULL)
	   
	   free(All_words_lists) ; 
       }
       
       // gets the IL of the token with number rank 
       res=XCDE_Get_IL_ListPosProx(rank,
				   &IL_Word, 
				   &IL_Word_num,
				   &XCDE_Console) ;
	   
       if (res == -1) { 

	 fprintf(stderr, "\nFatal Error: problems with the XCDE_Get_IL_ListPos function\n") ;
	 
	 fflush(stderr) ;
	 
	 if (tokendict != NULL)
	   
	   free(tokendict) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (All_words_lists_lengths != NULL) 
	   
	   free(All_words_lists_lengths) ; 
	 
	 for(k=0 ; k<wordcounter; ++k) 
	   
	   free(All_words_lists[k]) ; 
	 
	 if (All_words_lists != NULL)
	   
	   free(All_words_lists) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
       }
       
       // first select: takes only the words that are contained into the tags 
       if (tagcounter > 0) {
	 
	 res=select_words_contained_in_tags(&New_IL_Word, 
					    &New_IL_Word_num,
					    IL_Tags, 
					    IL_Tags_num,
					    IL_Word, 
					    IL_Word_num) ; 
       
       if (res == -1) 
	 
	 { fprintf(stderr, "\nFatal Error: problems with the select_words_contained_in_tags function\n") ;
	 
	 fflush(stderr) ;
	 
	 if (tokendict != NULL)
	   
	   free(tokendict) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	     
	 if (All_words_lists_lengths != NULL) 
	   
	   free(All_words_lists_lengths) ; 
	 
	 for(k=0 ; k<wordcounter; ++k) 
	       
	   free(All_words_lists[k]) ; 
	 
	 if (All_words_lists != NULL)
	   
	   free(All_words_lists) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
	 }
       
       // free the memory needed to contain the old list
       if (IL_Word != NULL) { 

	 free(IL_Word) ;
	 
	 IL_Word=NULL ;
       }
       
       // points to the new list 
       IL_Word=New_IL_Word ;
       
       IL_Word_num=New_IL_Word_num ; 
       
       New_IL_Word_num=0 ;
       
       New_IL_Word=NULL ; 
       }

       // join the words list  
       res=join_words(&New_IL_Word, 
		      &New_IL_Word_num, 
		      IL_Words,
		      IL_Words_num,
		      IL_Word,
		      IL_Word_num) ;
       
       if (res == -1) { 

	 fprintf(stderr, "\nFatal Error: problems with the join_words function\n") ;
	 
	 fflush(stderr) ;
	 
	 if (tokendict != NULL)
	   
	   free(tokendict) ; 
	 
	 if (IL_Tags != NULL)
	   
	   free(IL_Tags) ; 
	 
	 if (All_words_lists_lengths != NULL) 
	   
	   free(All_words_lists_lengths) ; 
	 
	 for(k=0 ; k<wordcounter; ++k) 
	   
	   free(All_words_lists[k]) ; 
	 
	 if (All_words_lists != NULL)
	   
	   free(All_words_lists) ; 
	 
	 XCDE_Finalize_Console(&XCDE_Console) ;
	 
	 exit(-1) ; 
	 }
       
       // frees the memory needed to contain the old lists
       if (IL_Words != NULL) { 

	 free(IL_Words) ;
	 
	 IL_Words=NULL ;
	 }

       // copies the new list 
       IL_Words=New_IL_Word ;

       IL_Words_num=New_IL_Word_num ;

       New_IL_Word=NULL ;
       
       New_IL_Word_num=0 ;
       
       // frees the memory needed to contain the old list 
       if (IL_Word != NULL) { 

	 free(IL_Word) ; 
	 
	 IL_Word=NULL ;
       }
       
       IL_Word_num=0 ; 
   }
   
   // frees the memory needed to contain the occurrencies in the dictionary 
   if (tokendict != NULL) { 
     
     free(tokendict) ;
     
     tokendict=NULL ;
   }  
   
   numtoken=0 ; 
   
   // checks if there're no occurrencies: in this case finalizes the console and ends
   if (IL_Words_num == 0) { 
     if (IL_Tags !=NULL)
       
     free(IL_Tags) ; 
   
   XCDE_Finalize_Console(&XCDE_Console) ;
   
   return(1) ;
   }

   
   if (All_words_dimension == 0) { // creates a matrix needed to use the plane-sweep algorithm 
     
     All_words_lists=(XCDE_IL_Triple_Type **) malloc(5*sizeof(XCDE_IL_Triple_Type *)) ;
       
     All_words_lists_lengths=(int *) malloc(5*sizeof(int)) ; 
     
     All_words_dimension=5 ;
   }

   else 
       
     if (wordcounter >= All_words_dimension) { 

       All_words_lists=(XCDE_IL_Triple_Type **) realloc(All_words_lists,
							All_words_dimension*2*sizeof(XCDE_IL_Triple_Type *)) ; 
       
       All_words_lists_lengths=(int *) realloc(All_words_lists_lengths,
					       All_words_dimension*2*sizeof(int)) ; 
       
       All_words_dimension*=2 ;
     }
   
   
   // copies the list of occurrencies of the word in the matrix used to executes the plane-sweep algorithm 
   All_words_lists[wordcounter-1]=IL_Words ; 
   
   All_words_lists_lengths[wordcounter-1]=IL_Words_num ; 
     
   IL_Words=NULL ; 
   
   IL_Words_num=0 ; 
 }

 // ..
 // ..  THE QUERY HAS BEEN ANALYZED. PRODUCES THE RESULT
 // ..  

 // checks if the number of words queries is greater than one. In this case applies the plane-sweep algorithm 
 if (wordcounter > 1) { // if enters this block there are more than one occurrence. Applies the
                        // plane-sweep algorithm 

   // the plane-sweep is an algorithm used to select all the occurrencies that have a setted proximity
   res=plane_sweep_proximity(wordcounter,
			     All_words_lists,
			     All_words_lists_lengths, 
			     &plane_sweep_proximity_windows_results,
			     &plane_sweep_proximity_windows_results_num,
			     proximity_num) ;
     
   if (res == -1) { 

     fprintf(stderr, "\nFatal Error: problems with the plane sweep proximity algorithm\n") ;
     
     fflush(stderr) ; 
     
     if (IL_Tags != NULL)
       
       free(IL_Tags) ;
     
     for(k=0 ; k<wordcounter; ++k)
       
       if (All_words_lists[k] != NULL)
	 
	 free(All_words_lists[k]) ;
     
     if (All_words_lists != NULL)
       
       free(All_words_lists) ; 
     
     if (plane_sweep_proximity_windows_results != NULL)
       
       free(plane_sweep_proximity_windows_results) ; 
     
     XCDE_Finalize_Console(&XCDE_Console) ; 
     
       return(-1) ;
   }

   // selects only the proximity windows obtained from the previous algorithm that're contained
   // in the tags occurrencies 
   if (tagcounter > 0) 

     { res=select_proximity_windows_contained_in_tags(&New_plane_sweep_proximity_windows_results, 
						      &New_plane_sweep_proximity_windows_results_num, 
						      IL_Tags, 
						      IL_Tags_num, 
						      plane_sweep_proximity_windows_results,
						      plane_sweep_proximity_windows_results_num) ; 
     
     if (res == -1) { 
       
       fprintf(stderr, "\nFatal Error: problems with the select_proximity_windows_contained_in_tags function\n") ;
       
       fflush(stderr) ; 
     
       if (All_words_lists_lengths != NULL)
       
       free(All_words_lists_lengths) ;
     
       for(k=0 ; k<wordcounter; ++k)
       
	 if (All_words_lists[k] != NULL)
	 
	   free(All_words_lists[k]) ;
       
       if (All_words_lists != NULL)
	 
	 free(All_words_lists) ; 
       
       if (plane_sweep_proximity_windows_results != NULL)
	 
	 free(plane_sweep_proximity_windows_results) ; 
       
       if (New_plane_sweep_proximity_windows_results != NULL)
	 
	 free(New_plane_sweep_proximity_windows_results) ; 
       
       XCDE_Finalize_Console(&XCDE_Console) ; 
       
       return(-1) ;
     }
   
   // frees the memory needed to contain old proximity lists 
   if (plane_sweep_proximity_windows_results != NULL) { 
     
     free(plane_sweep_proximity_windows_results) ;
     
     plane_sweep_proximity_windows_results=NULL ; 
   }
     }
   // copies the new proximity list
   plane_sweep_proximity_windows_results=New_plane_sweep_proximity_windows_results ;

   plane_sweep_proximity_windows_results_num=New_plane_sweep_proximity_windows_results_num ; 

   New_plane_sweep_proximity_windows_results=0 ;
   
   New_plane_sweep_proximity_windows_results_num=0 ; 

   // prints the resulting windows : namefile, start_position, ending_position for each occurrence
   for(k=0 ; k<plane_sweep_proximity_windows_results_num; k+=2) 

     fprintf(stdout, "%s %d %d\n", argv[1], plane_sweep_proximity_windows_results[k].XCDE_Position, plane_sweep_proximity_windows_results[k+1].XCDE_Position) ;
 }
 
 else 

   // checks if there's only one word query 
   if (wordcounter == 1)
   
     // if executes the for command there's only one word query, so the execution of plane-sweep algorithm
     // isn't necessary. Prints the resulting windows, that are the occurrence of the only word
     for (k=0 ; k<All_words_lists_lengths[0] ; ++k) 
     
       fprintf(stdout, "%s %d %d\n", argv[1], All_words_lists[0][k].XCDE_Position, All_words_lists[0][k].XCDE_Position) ;
 
   else
     
     // checks if there're no word queries 
     for(k=0 ; k<IL_Tags_num; k+=2) // if executes the for command there're no word queries. Prints the resulting windows
                                    // that are all the positions in the list computed for the tags 
       
       fprintf(stdout, "%s %d %d\n", argv[1], IL_Tags[k].XCDE_Position, IL_Tags[k+1].XCDE_Position) ;
 
 // finalizes the console
 XCDE_Finalize_Console(&XCDE_Console) ; 

 // frees the memory used
 if (plane_sweep_proximity_windows_results != NULL) 

   free(plane_sweep_proximity_windows_results) ; 

 if (All_words_lists_lengths != NULL)
   
   free(All_words_lists_lengths) ;
 
 for(k=0 ; k<wordcounter; ++k)
   
   if (All_words_lists[k] != NULL)
     
     free(All_words_lists[k]) ;
 
 if (All_words_lists != NULL)
   
   free(All_words_lists) ;
 
 if (IL_Tags != NULL)
   
   free(IL_Tags) ;
 
 return(0) ;
}


// function used to join lists of positions of attributes or words and
// created a new sorted list that contains all the position   
int join_attrs_or_words(XCDE_IL_Couple_Type **IL_new, 
			int *IL_new_num_el,
			XCDE_IL_Couple_Type *IL1,
			int IL1_num_el,
			XCDE_IL_Couple_Type *IL2,
			int IL2_num_el)
{ int pointer_IL1=0 ;

 int pointer_IL2=0 ; 

 int pointer_IL_new=0 ;
 
 // checks parameters
 if (IL_new == NULL)

   return(-1) ;
  
 // allocates the memory needed for the new list 
 (*IL_new)=(XCDE_IL_Couple_Type *) malloc((IL1_num_el+IL2_num_el)*sizeof(XCDE_IL_Couple_Type)) ;

 if (*IL_new == NULL)
   
   return(-1) ;

 // loops over all the positions and copies them in order 
 while ((pointer_IL1 < IL1_num_el) && (pointer_IL2 < IL2_num_el)) 

   if ((IL1[pointer_IL1].XCDE_Position < IL2[pointer_IL2].XCDE_Position) ||
       ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	(IL1[pointer_IL1].XCDE_Offset < IL2[pointer_IL2].XCDE_Offset))) { // the position of the first list 
                                                                          // must be copied

     (*IL_new)[pointer_IL_new].XCDE_Position = IL1[pointer_IL1].XCDE_Position ;

     (*IL_new)[pointer_IL_new].XCDE_Offset = IL1[pointer_IL1].XCDE_Offset ;
     
     ++pointer_IL_new ;

     ++pointer_IL1 ;
   }

   else
     
     if ((IL1[pointer_IL1].XCDE_Position > IL2[pointer_IL2].XCDE_Position) ||
	 ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	  (IL1[pointer_IL1].XCDE_Offset > IL2[pointer_IL2].XCDE_Offset))) { // the position of the second list 
                                                                            // must be copied 

       (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
       
       (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;
       
       ++pointer_IL_new ;
       
       ++pointer_IL2 ;
     }

     else
   
       if  ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	    (IL1[pointer_IL1].XCDE_Offset == IL2[pointer_IL2].XCDE_Offset)) { // the position of the first and second
	                                                                      // list (they are the same) must be copied

	 (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
	 
	 (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;
	 
	 ++pointer_IL_new ;
	 
	 ++pointer_IL2 ;

	 ++pointer_IL1 ;
	 }

 // copies the element of the first list 
 while (pointer_IL1 < IL1_num_el) { 

   (*IL_new)[pointer_IL_new].XCDE_Position = IL1[pointer_IL1].XCDE_Position ;
     
   (*IL_new)[pointer_IL_new].XCDE_Offset = IL1[pointer_IL1].XCDE_Offset ;
   
   ++pointer_IL_new ;
   
   ++pointer_IL1 ;
 }
 
 // copies the element of the second list 
 while (pointer_IL2 < IL2_num_el)

   { (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
   
   (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;
     
   ++pointer_IL_new ;

   ++pointer_IL2 ;
   }

 // copies the new list 
 (*IL_new_num_el)=pointer_IL_new ;

 return(1) ;
}

// function used to join lists of positions of words 
int join_words(XCDE_IL_Triple_Type **IL_new, 
	       int *IL_new_num_el,
	       XCDE_IL_Triple_Type *IL1,
	       int IL1_num_el,
	       XCDE_IL_Triple_Type *IL2,
	       int IL2_num_el)
{ int pointer_IL1=0 ;

 int pointer_IL2=0 ; 

 int pointer_IL_new=0 ;
 
 // checks parameters
 if (IL_new == NULL)

   return(-1) ;
  
 // allocates the memory needed to contain the new list
 (*IL_new)=(XCDE_IL_Triple_Type *) malloc((IL1_num_el+IL2_num_el)*sizeof(XCDE_IL_Triple_Type)) ;

 if (*IL_new == NULL)

   return(-1) ;

 // loops over the two list 
 while ((pointer_IL1 < IL1_num_el) && (pointer_IL2 < IL2_num_el)) 

   // checks if the occurrence in the first list is before the occurrence in the second one 
   if ((IL1[pointer_IL1].XCDE_Position < IL2[pointer_IL2].XCDE_Position) ||
       ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	(IL1[pointer_IL1].XCDE_Offset < IL2[pointer_IL2].XCDE_Offset))) { // the occurrence in the first list is before the
                                                                          // occurrence in the second one. Copies the first list occurrence

     // copies the occurrence in the first list 
     (*IL_new)[pointer_IL_new].XCDE_Position = IL1[pointer_IL1].XCDE_Position ;

     (*IL_new)[pointer_IL_new].XCDE_Offset = IL1[pointer_IL1].XCDE_Offset ;
     
     (*IL_new)[pointer_IL_new].XCDE_Proximity= IL1[pointer_IL1].XCDE_Proximity ;

     // increases the new list pointer
     ++pointer_IL_new ;

     // increases the first list pointer 
     ++pointer_IL1 ;
     }

   else
     
     // checks if the occurrence in the second list is before the occurrence in the first one. 
     if ((IL1[pointer_IL1].XCDE_Position > IL2[pointer_IL2].XCDE_Position) ||
	 ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	  (IL1[pointer_IL1].XCDE_Offset > IL2[pointer_IL2].XCDE_Offset))){  // If enters this block the occurrence in the second list is 
                                                                            // before the occurrence in the first one. Copies the second list occurrence
       
       // copies the second list occurrence
       (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
       
       (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;
       
       (*IL_new)[pointer_IL_new].XCDE_Proximity= IL2[pointer_IL2].XCDE_Proximity ;

       // increases the new list pointer
       ++pointer_IL_new ;

       // increases the second list pointer
       ++pointer_IL2 ;
     }

     else
   
       // the occurrence of the first list and of the second list are the same
       if  ((IL1[pointer_IL1].XCDE_Position == IL2[pointer_IL2].XCDE_Position) &&
	    (IL1[pointer_IL1].XCDE_Offset == IL2[pointer_IL2].XCDE_Offset)) { // If enters this block the occurrence of the first list and 
	                                                                      // of the second list are the same. Copies the occurrence in the
	                                                                      // second list and increases all the pointers 

	 // copies the occurrence in the second list 
	 (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
	 
	 (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;
	 
	 (*IL_new)[pointer_IL_new].XCDE_Proximity= IL2[pointer_IL2].XCDE_Proximity ;

	 // increases the new list pointer
	 ++pointer_IL_new ;
	 
	 // increases the second list pointer
	 ++pointer_IL2 ;

	 // increases the first list pointer 
	 ++pointer_IL1 ;
       }

 // checks if the first list is smaller than the second one
 while (pointer_IL1 < IL1_num_el) { // if enters in the loop the second list is smaller than the second first one. Copies all the 
                                    // elements of the first list not alyet copied, that are greater than the elements of the second list

   // copies the element in the first list 
   (*IL_new)[pointer_IL_new].XCDE_Position = IL1[pointer_IL1].XCDE_Position ;
     
   (*IL_new)[pointer_IL_new].XCDE_Offset = IL1[pointer_IL1].XCDE_Offset ;
     
   (*IL_new)[pointer_IL_new].XCDE_Proximity= IL1[pointer_IL1].XCDE_Proximity ;
   
   // increases the new list pointer
   ++pointer_IL_new ;
   
   // increases the first list pointer 
   ++pointer_IL1 ;
 }
 
 // checks if the second list is smaller than the first one
 while (pointer_IL2 < IL2_num_el) { // if enters in the loop the first list is smaller than the first one. Copies all the elements
                                    // of the second list not alyet copied, that are greater than the elements of the first list 

   // copies the element in the second list
   (*IL_new)[pointer_IL_new].XCDE_Position = IL2[pointer_IL2].XCDE_Position ;
   
   (*IL_new)[pointer_IL_new].XCDE_Offset = IL2[pointer_IL2].XCDE_Offset ;

   (*IL_new)[pointer_IL_new].XCDE_Proximity= IL2[pointer_IL2].XCDE_Proximity ;
     
   // increases the new list pointer
   ++pointer_IL_new ;

   // increases the second list pointer
   ++pointer_IL2 ;
   }
 
 // copies the new list 
 (*IL_new_num_el)=pointer_IL_new ;

 return(1) ;
}

// functions used to select all occurrences of tags that contains one 
// or more attributes
int select_tag_with_attrs(XCDE_IL_Couple_Type **selected_IL,
			  int *num_el_selected_IL,
			  XCDE_IL_Couple_Type *tag_IL,
			  int num_el_tag_IL,
			  XCDE_IL_Couple_Type *attr_IL,
			  int num_el_attr_IL) 
{ int pointer_selected_IL=0 ;

  int pointer_tag_IL=0 ;

  int pointer_attr_IL=0 ;
 
  // checks if the parameter is correct 
  if (selected_IL == NULL) 

  return(-1) ;

  // allocates the memory needed to contain the tag list 
  *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL)*sizeof(XCDE_IL_Triple_Type)) ;

  while((pointer_tag_IL < num_el_tag_IL) && (pointer_attr_IL < num_el_attr_IL)) 

    if ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	(tag_IL[pointer_tag_IL].XCDE_Offset == attr_IL[pointer_attr_IL].XCDE_Offset)) { // the attributes is contained by the tag occurrence. 
                                                                                        // Copies it

      (*selected_IL)[pointer_selected_IL].XCDE_Position=tag_IL[pointer_tag_IL].XCDE_Position ;
      
      (*selected_IL)[pointer_selected_IL].XCDE_Offset=tag_IL[pointer_tag_IL].XCDE_Offset ;
      
      (*selected_IL)[pointer_selected_IL+1].XCDE_Position=tag_IL[pointer_tag_IL+1].XCDE_Position ;
      
      (*selected_IL)[pointer_selected_IL+1].XCDE_Offset=tag_IL[pointer_tag_IL+1].XCDE_Offset ;
      
      pointer_selected_IL+=2 ;

      // increases the position of the tag
      pointer_tag_IL+=2 ;
      
      // increases the position of the attribute IL 
      ++pointer_attr_IL ; 
    }
  
    else
      
      if ((tag_IL[pointer_tag_IL].XCDE_Position < attr_IL[pointer_attr_IL].XCDE_Position) ||
	  ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	   (tag_IL[pointer_tag_IL].XCDE_Offset < attr_IL[pointer_attr_IL].XCDE_Offset))) // the attribute position is greater than the tag position. Increases it 
	
	// increases the position of the tag IL 
	pointer_tag_IL+=2 ;
  
      else
	
	if ((tag_IL[pointer_tag_IL].XCDE_Position > attr_IL[pointer_attr_IL].XCDE_Position) ||
	    ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	     (tag_IL[pointer_tag_IL].XCDE_Offset > attr_IL[pointer_attr_IL].XCDE_Offset))) // the attribute position is less that the tag position. Increases the first
	  
	  ++pointer_attr_IL ; 
  
  // copies the list 
  *num_el_selected_IL=pointer_selected_IL ;

  return(1) ;
}

// function used to select all occurrencies of a tag that doesn't contain 
// one or more attributes 
int select_tag_without_attrs(XCDE_IL_Couple_Type **selected_IL,
			     int *num_el_selected_IL,
			     XCDE_IL_Couple_Type *tag_IL,
			     int num_el_tag_IL,
			     XCDE_IL_Couple_Type *attr_IL,
			     int num_el_attr_IL) 
{ int pointer_selected_IL=0 ;

  int pointer_tag_IL=0 ;

  int pointer_attr_IL=0 ;
 
  // checks if the parameters are correct
  if (selected_IL == NULL) 

  return(-1) ;

  // allocates the memory needed to contain the new list 
  *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL)*sizeof(XCDE_IL_Couple_Type)) ;

  // loops over the two lists (tag and attributes)
  while((pointer_tag_IL < num_el_tag_IL) && (pointer_attr_IL < num_el_attr_IL)) 

    if ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	(tag_IL[pointer_tag_IL].XCDE_Offset == attr_IL[pointer_attr_IL].XCDE_Offset)) { // the opening position of the tag is equal to the 
                                                                                        // position of the attribute, so the tag position isn't copied

      // increases the position of the tag list
      pointer_tag_IL+=2 ;

      // increases the position of the attribute list 
      ++pointer_attr_IL ;
    }

    else

      if ((tag_IL[pointer_tag_IL].XCDE_Position < attr_IL[pointer_attr_IL].XCDE_Position) ||
	  ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	   (tag_IL[pointer_tag_IL].XCDE_Offset < attr_IL[pointer_attr_IL].XCDE_Offset))) { // the tag dosn't contain the attribute, so its positions of opening
	                                                                                   // and closing are copied 

	// copies the opening and closing position of the tag 
	(*selected_IL)[pointer_selected_IL].XCDE_Position=tag_IL[pointer_tag_IL].XCDE_Position ;
      
	(*selected_IL)[pointer_selected_IL].XCDE_Offset=tag_IL[pointer_tag_IL].XCDE_Offset ;

	(*selected_IL)[pointer_selected_IL+1].XCDE_Position=tag_IL[pointer_tag_IL+1].XCDE_Position ;
	
	(*selected_IL)[pointer_selected_IL+1].XCDE_Offset=tag_IL[pointer_tag_IL+1].XCDE_Offset ;
	
	// increases the position of the attribute IL 
	pointer_tag_IL+=2 ;

	// increases the position of the TAG
	pointer_selected_IL+=2 ;
	}

      else
	
	// the attribute position is less that the tag position. Increases the attribute list
	if ((tag_IL[pointer_tag_IL].XCDE_Position > attr_IL[pointer_attr_IL].XCDE_Position) ||
	    ((tag_IL[pointer_tag_IL].XCDE_Position == attr_IL[pointer_attr_IL].XCDE_Position) &&
	     (tag_IL[pointer_tag_IL].XCDE_Offset > attr_IL[pointer_attr_IL].XCDE_Offset)))
	  
	  
	  ++pointer_attr_IL ; 
  
  // copies the TAG occurrencies that doesn't contain the attribute 
  while (pointer_tag_IL < num_el_tag_IL) { 

    (*selected_IL)[pointer_selected_IL].XCDE_Position=tag_IL[pointer_tag_IL].XCDE_Position ;
    
    (*selected_IL)[pointer_selected_IL].XCDE_Offset=tag_IL[pointer_tag_IL].XCDE_Offset ;
    
    (*selected_IL)[pointer_selected_IL+1].XCDE_Position=tag_IL[pointer_tag_IL+1].XCDE_Position ;
    
    (*selected_IL)[pointer_selected_IL+1].XCDE_Offset=tag_IL[pointer_tag_IL+1].XCDE_Offset ;
    
    pointer_selected_IL+=2 ;
    
    pointer_tag_IL+=2 ;
  }
  
  *num_el_selected_IL=pointer_selected_IL ;
  
  return(1) ;
}

// function used to select all occurrencies of a tag that 
// are nested in other tag
int select_tag_nested_in_tag(XCDE_IL_Couple_Type **selected_IL,
			     int *num_el_selected_IL,
			     XCDE_IL_Couple_Type *tag_IL1,
			     int num_el_tag_IL1,
			     XCDE_IL_Couple_Type *tag_IL2,
			     int num_el_tag_IL2) 
{ int pointer_selected_IL=0 ;
 
 int pointer_tag_IL1=0 ;
 
 int pointer_tag_IL2=0 ; 
 
 // allocates the memory needed to contain the list 
 *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL2)*sizeof(XCDE_IL_Couple_Type)) ;

 // loops over the two lists
 while((pointer_tag_IL1 < num_el_tag_IL1) && (pointer_tag_IL2 < num_el_tag_IL2)) 
   
   // checks if the first tag position contains the second one
    if (((tag_IL1[pointer_tag_IL1].XCDE_Position < tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1].XCDE_Offset < tag_IL2[pointer_tag_IL2].XCDE_Offset))) &&
	(tag_IL1[pointer_tag_IL1+1].XCDE_Position > tag_IL2[pointer_tag_IL2+1].XCDE_Position) ||
	((tag_IL1[pointer_tag_IL1+1].XCDE_Position == tag_IL2[pointer_tag_IL2+1].XCDE_Position) &&
	 (tag_IL1[pointer_tag_IL1+1].XCDE_Offset > tag_IL2[pointer_tag_IL2+1].XCDE_Offset))) { // if enters this block the first tag position contains
                                                                                               // the second one, so copies it in the new list  

	(*selected_IL)[pointer_selected_IL].XCDE_Position = tag_IL2[pointer_tag_IL2].XCDE_Position ;
      
	(*selected_IL)[pointer_selected_IL].XCDE_Offset = tag_IL2[pointer_tag_IL2].XCDE_Offset ;
	
	(*selected_IL)[pointer_selected_IL+1].XCDE_Position = tag_IL2[pointer_tag_IL2+1].XCDE_Position ;
	
	(*selected_IL)[pointer_selected_IL+1].XCDE_Offset = tag_IL2[pointer_tag_IL2+1].XCDE_Offset ;
	
	// increases the pointer to the list 
	pointer_selected_IL+=2 ;
	
	pointer_tag_IL2+=2 ;
    }
    
    else

      // checks if the first list position is greater than the second one 
      if ((tag_IL1[pointer_tag_IL1].XCDE_Position >= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	  ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	   (tag_IL1[pointer_tag_IL1].XCDE_Offset >= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	
	// in this case the first list position is greater than the second one. Increases the second position list 
	pointer_tag_IL2+=2 ;

      else
	
	// checks if the second list position is greater than the first one 
	if ((tag_IL1[pointer_tag_IL1].XCDE_Position <= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	    ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	     (tag_IL1[pointer_tag_IL1].XCDE_Offset <= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	  
	  // in this case the second list position is greater than the first one. Increases the first position list 
	  pointer_tag_IL1+=2 ;
	      
 // copies the list  
 *num_el_selected_IL=pointer_selected_IL ;

 return(1) ;
}
	      
// function used to select all occurrencies of a tag that
// are nested in other tag with the specified distance
int select_tag_nested_in_tag_with_distance(XCDE_IL_Couple_Type **selected_IL,
					   int *num_el_selected_IL,
					   int rank, 
					   int dist,
					   XCDE_IL_Couple_Type *tag_IL1,
					   int num_el_tag_IL1,
					   XCDE_IL_Couple_Type *tag_IL2,
					   int num_el_tag_IL2,
					   XCDE_Console_Type *Console) 
{ int res=0 ;

  int pointer_selected_IL=0 ;

  int pointer_tag_IL1=0 ;

  int pointer_tag_IL2=0 ; 

  int *enclosing=NULL ;
  
  int num_enclosing=0 ;

  // allocates the memory needed to contain the tag 
  *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL2)*sizeof(XCDE_IL_Couple_Type)) ;
  
  // loops over lists of positions of the tags 
  while((pointer_tag_IL1 < num_el_tag_IL1) && (pointer_tag_IL2 < num_el_tag_IL2)) { 

    if (((tag_IL1[pointer_tag_IL1].XCDE_Position < tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1].XCDE_Offset < tag_IL2[pointer_tag_IL2].XCDE_Offset))) &&
	(tag_IL1[pointer_tag_IL1+1].XCDE_Position > tag_IL2[pointer_tag_IL2+1].XCDE_Position) ||
	((tag_IL1[pointer_tag_IL1+1].XCDE_Position == tag_IL2[pointer_tag_IL2+1].XCDE_Position) &&
	 (tag_IL1[pointer_tag_IL1+1].XCDE_Offset > tag_IL2[pointer_tag_IL2+1].XCDE_Offset))) { 

      
      // the second tag position is contained in the first tag position. So use the tagstract of the second position 
      // to check the distance
      res=XCDE_Get_Tagstruct_Tags_Enclosing(&tag_IL2[pointer_tag_IL2],
					    &enclosing,
					    &num_enclosing,
					    Console) ;

      if (res == -1) { fprintf(stderr, "\nFatal Error: problems with the XCDE_Get_Tagstruct_Tags_Enclosing function\n") ;
      
      fflush(stderr) ;
      
      XCDE_Finalize_Console(Console) ;
      
      exit(-1) ;
      }
      
      // checks if the first tag is at distance dist from the second 
      if (((num_enclosing-1) >= dist) && (enclosing[num_enclosing-(dist+1)] == rank)) { // if enters this block the first tag is at distance dist
	                                                                                // from the second. So copies it.

	(*selected_IL)[pointer_selected_IL].XCDE_Position = tag_IL2[pointer_tag_IL2].XCDE_Position ;
	
	(*selected_IL)[pointer_selected_IL].XCDE_Offset = tag_IL2[pointer_tag_IL2].XCDE_Offset ;
	
	(*selected_IL)[pointer_selected_IL+1].XCDE_Position = tag_IL2[pointer_tag_IL2+1].XCDE_Position ;
	
	(*selected_IL)[pointer_selected_IL+1].XCDE_Offset = tag_IL2[pointer_tag_IL2+1].XCDE_Offset ;
	
	pointer_selected_IL+=2 ;
      }

      // frees the memory needed to contain the tagstruct 
      if (enclosing != NULL) { 
	
	free(enclosing) ;
	
	enclosing=NULL ;
      }
      
      num_enclosing=0 ;
      
      pointer_tag_IL2+=2 ;
    }
        
    else
      
      if ((tag_IL1[pointer_tag_IL1].XCDE_Position >= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	  ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	   (tag_IL1[pointer_tag_IL1].XCDE_Offset >= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	
	// increases the second tag list position because it's less that the first 
	pointer_tag_IL2+=2 ;

      else
	
	if ((tag_IL1[pointer_tag_IL1].XCDE_Position <= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	    ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	     (tag_IL1[pointer_tag_IL1].XCDE_Offset <= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	  
	  // increases the first tag list position because it's less that the second 
	  pointer_tag_IL1+=2 ;
    }
  
  // copies the list obtained 
  *num_el_selected_IL=pointer_selected_IL ;

  return(1) ;
}

int prune_tag_with_word(XCDE_IL_Couple_Type **selected_IL,
			int *num_el_selected_IL,
			XCDE_IL_Couple_Type *tag_IL1,
			int num_el_tag_IL1,
			XCDE_IL_Couple_Type *tag_IL2,
			int num_el_tag_IL2) 
{ int pointer_selected_IL=0 ;

  int pointer_tag_IL1=0 ;

  int pointer_tag_IL2=0 ; 

  *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL1)*sizeof(XCDE_IL_Couple_Type)) ;

  while((pointer_tag_IL1 < num_el_tag_IL1) && (pointer_tag_IL2 < num_el_tag_IL2)) 

    if (((tag_IL1[pointer_tag_IL1].XCDE_Position < tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1].XCDE_Offset < tag_IL2[pointer_tag_IL2].XCDE_Offset))) &&
	(tag_IL1[pointer_tag_IL1+1].XCDE_Position > tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1+1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1+1].XCDE_Offset > tag_IL2[pointer_tag_IL2].XCDE_Offset)))

      { (*selected_IL)[pointer_selected_IL].XCDE_Position = tag_IL1[pointer_tag_IL1].XCDE_Position ;

	(*selected_IL)[pointer_selected_IL].XCDE_Offset = tag_IL1[pointer_tag_IL1].XCDE_Offset ;

	(*selected_IL)[pointer_selected_IL+1].XCDE_Position = tag_IL1[pointer_tag_IL1+1].XCDE_Position ;

	(*selected_IL)[pointer_selected_IL+1].XCDE_Offset = tag_IL1[pointer_tag_IL1+1].XCDE_Offset ;

	pointer_selected_IL+=2 ;

	pointer_tag_IL1+=2 ;
      }
    
    else

      if ((tag_IL1[pointer_tag_IL1].XCDE_Position >= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	  ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	   (tag_IL1[pointer_tag_IL1].XCDE_Offset >= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	
	++pointer_tag_IL2 ;

      else
	
	if ((tag_IL1[pointer_tag_IL1].XCDE_Position <= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	    ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	     (tag_IL1[pointer_tag_IL1].XCDE_Offset <= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	  
	  pointer_tag_IL1+=2 ;
	      
  *num_el_selected_IL=pointer_selected_IL ;

  return(1) ;
}
	      
int prune_tag_with_word_prox(XCDE_IL_Couple_Type **selected_IL,
			     int *num_el_selected_IL,
			     XCDE_IL_Couple_Type *tag_IL1,
			     int num_el_tag_IL1,
			     XCDE_IL_Triple_Type *tag_IL2,
			     int num_el_tag_IL2) 	      
{ int pointer_selected_IL=0 ;

  int pointer_tag_IL1=0 ;

  int pointer_tag_IL2=0 ; 

  *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_tag_IL1)*sizeof(XCDE_IL_Couple_Type)) ;

  while((pointer_tag_IL1 < num_el_tag_IL1) && (pointer_tag_IL2 < num_el_tag_IL2)) 

    if (((tag_IL1[pointer_tag_IL1].XCDE_Position < tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1].XCDE_Offset < tag_IL2[pointer_tag_IL2].XCDE_Offset))) &&
	(tag_IL1[pointer_tag_IL1+1].XCDE_Position > tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	 ((tag_IL1[pointer_tag_IL1+1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	  (tag_IL1[pointer_tag_IL1+1].XCDE_Offset > tag_IL2[pointer_tag_IL2].XCDE_Offset)))

      { (*selected_IL)[pointer_selected_IL].XCDE_Position = tag_IL1[pointer_tag_IL1].XCDE_Position ;

	(*selected_IL)[pointer_selected_IL].XCDE_Offset = tag_IL1[pointer_tag_IL1].XCDE_Offset ;

	(*selected_IL)[pointer_selected_IL+1].XCDE_Position = tag_IL1[pointer_tag_IL1+1].XCDE_Position ;

	(*selected_IL)[pointer_selected_IL+1].XCDE_Offset = tag_IL1[pointer_tag_IL1+1].XCDE_Offset ;

	pointer_selected_IL+=2 ;

	pointer_tag_IL1+=2 ;
      }
    
    else

      if ((tag_IL1[pointer_tag_IL1].XCDE_Position >= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	  ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	   (tag_IL1[pointer_tag_IL1].XCDE_Offset >= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	
	++pointer_tag_IL2 ;

      else
	
	if ((tag_IL1[pointer_tag_IL1].XCDE_Position <= tag_IL2[pointer_tag_IL2].XCDE_Position) ||
	    ((tag_IL1[pointer_tag_IL1].XCDE_Position == tag_IL2[pointer_tag_IL2].XCDE_Position) &&
	     (tag_IL1[pointer_tag_IL1].XCDE_Offset <= tag_IL2[pointer_tag_IL2].XCDE_Offset)))
	  
	  pointer_tag_IL1+=2 ;
	      
  *num_el_selected_IL=pointer_selected_IL ;

  return(1) ;
}	  

// function used to select all occurencies of a word
// that are contained in a tag 
int select_words_contained_in_tags(XCDE_IL_Triple_Type **selected_IL, 
				   int *num_el_selected_IL, 
				   XCDE_IL_Couple_Type *tag_IL, 
				   int num_el_tag_IL, 
				   XCDE_IL_Triple_Type *word_IL,
				   int num_el_word_IL)
{ int pointer_selected_IL=0 ;

 int pointer_tag_IL=0 ;

 int pointer_word_IL=0 ; 

 // allocates the memory needed to contain the word occurrencies 
 *selected_IL = (XCDE_IL_Triple_Type *) malloc((num_el_word_IL)*sizeof(XCDE_IL_Triple_Type)) ;

 // loops over the two lists 
 while((pointer_tag_IL < num_el_tag_IL) && (pointer_word_IL < num_el_word_IL)) 

   // checks if the word occurrence is contained in the tag occurrence
   if (((tag_IL[pointer_tag_IL].XCDE_Position < word_IL[pointer_word_IL].XCDE_Position) ||
	((tag_IL[pointer_tag_IL].XCDE_Position == word_IL[pointer_word_IL].XCDE_Position) &&
	 (tag_IL[pointer_tag_IL].XCDE_Offset < word_IL[pointer_word_IL].XCDE_Offset))) &&
       (tag_IL[pointer_tag_IL+1].XCDE_Position > word_IL[pointer_word_IL].XCDE_Position) ||
       ((tag_IL[pointer_tag_IL+1].XCDE_Position == word_IL[pointer_word_IL].XCDE_Position) &&
	(tag_IL[pointer_tag_IL+1].XCDE_Offset > word_IL[pointer_word_IL].XCDE_Offset))) { // if enters this block the word occurrence is contained 
                                                                                          // in the tag, so copies the word occurrence

     
     // copies the word occurrence
     (*selected_IL)[pointer_selected_IL].XCDE_Position = word_IL[pointer_word_IL].XCDE_Position ;
     
     (*selected_IL)[pointer_selected_IL].XCDE_Offset = word_IL[pointer_word_IL].XCDE_Offset ;
     
     (*selected_IL)[pointer_selected_IL].XCDE_Proximity = word_IL[pointer_word_IL].XCDE_Proximity ;


     // increases the pointer to the new list
     ++pointer_selected_IL ;
	
     // increases the pointer to the tag list
     ++pointer_word_IL ;
   }
   
 
   else
     
     // checks if the word occurrence is before the tag occurrence. 
     if ((tag_IL[pointer_tag_IL].XCDE_Position >= word_IL[pointer_word_IL].XCDE_Position) ||
	 ((tag_IL[pointer_tag_IL].XCDE_Position == word_IL[pointer_word_IL].XCDE_Position) &&
	  (tag_IL[pointer_tag_IL].XCDE_Offset >= word_IL[pointer_word_IL].XCDE_Offset)))
       
       // the word occurrence is before the tag occurrence, increases the word list pointer
	++pointer_word_IL ;

      else
	
	// checks if the word occurrence is after the tag occurrence
	if ((tag_IL[pointer_tag_IL].XCDE_Position <= word_IL[pointer_word_IL].XCDE_Position) ||
	    ((tag_IL[pointer_tag_IL].XCDE_Position == word_IL[pointer_word_IL].XCDE_Position) &&
	     (tag_IL[pointer_tag_IL].XCDE_Offset <= word_IL[pointer_word_IL].XCDE_Offset)))
	  
	  // the word occurrence is after the tag occurrence, increases the word occurrence
	  pointer_tag_IL+=2 ;
	      
 // copies the list obtained
 *num_el_selected_IL=pointer_selected_IL ;
       
 return(1) ;
}	 

// function used to compute the proximity windows with the plane
// sweep algorithm
int plane_sweep_proximity(int number_proximal_lists, 
			  XCDE_IL_Triple_Type **Proximal_lists,
			  int *Proximal_Lists_Length,
			  XCDE_IL_Couple_Type **windows_result,
			  int *number_windows_result,
			  int proximity) 
{ register
    int k=0 ;
 
 unsigned
   char done=0 ;

 int *position=NULL ;
 
 int leftmost_element=0 ;

 int rightmost_element=0 ;

 int leftmost_list_position=0 ;

 int rightmost_list_position=0 ;

 int result_list_dimension=0 ;

 // allocates the memory needed for the list pointers 
 position=(int *) calloc(number_proximal_lists,
			 sizeof(int)) ;

 result_list_dimension=20 ;

 // allocates the memory for result windows
 *windows_result=(XCDE_IL_Couple_Type *) malloc(result_list_dimension*sizeof(XCDE_IL_Couple_Type)) ;

 *number_windows_result=0 ;

 while(!done) 

   { leftmost_element=INT_MAX ;

     rightmost_element=-1 ;

     // find leftmost and rightmost keyword, and their positions in the lists
     for (k=0 ; k<number_proximal_lists; ++k)
   
       { if (Proximal_lists[k][position[k]].XCDE_Proximity < leftmost_element)
	 
	 { leftmost_element=Proximal_lists[k][position[k]].XCDE_Proximity ;
	 
	 leftmost_list_position=k ;
	 }

       if (Proximal_lists[k][position[k]].XCDE_Proximity > rightmost_element)
	 
	 { rightmost_element= Proximal_lists[k][position[k]].XCDE_Proximity ;
	 
	 rightmost_list_position=k ;
	 }
       }
     
     if ((position[leftmost_list_position]+1) >= Proximal_Lists_Length[leftmost_list_position])

       { if ((proximity == 0) || ((proximity > 0) && ((rightmost_element-leftmost_element) <= proximity)))
	   
	   { (*windows_result)[(*number_windows_result)].XCDE_Position = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Position ;
	   
	   (*windows_result)[(*number_windows_result)].XCDE_Offset = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Offset ;
	   
	   (*windows_result)[(*number_windows_result+1)].XCDE_Position = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Position ;
	   
	   (*windows_result)[(*number_windows_result+1)].XCDE_Offset = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Offset ;
	   
	   (*number_windows_result)+=2 ;
	   }
	 
	 if ((*number_windows_result) >= result_list_dimension)
	   
	   { *windows_result=(XCDE_IL_Couple_Type *) realloc(*windows_result, 
							     result_list_dimension*2*sizeof(XCDE_IL_Couple_Type)) ;
	   
	   result_list_dimension*=2 ;
	   }
	 done=1 ; 
       }

     else

       if (Proximal_lists[leftmost_list_position][position[leftmost_list_position]+1].XCDE_Proximity > rightmost_element)
	 
	 // interval [l, r] is minimal, and we insert it in the result ; 
	 { if ((proximity == 0) || ((proximity > 0) && ((rightmost_element-leftmost_element) <= proximity)))
	   
	   { (*windows_result)[(*number_windows_result)].XCDE_Position = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Position ;
	   
	   (*windows_result)[(*number_windows_result)].XCDE_Offset = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Offset ;
	   
	   (*windows_result)[(*number_windows_result+1)].XCDE_Position = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Position ;
	   
	   (*windows_result)[(*number_windows_result+1)].XCDE_Offset = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Offset ;
	   
	   (*number_windows_result)+=2 ;
	   }
	 
	 if ((*number_windows_result) >= result_list_dimension)
	   
	   { *windows_result=(XCDE_IL_Couple_Type *) realloc(*windows_result, 
							     result_list_dimension*2*sizeof(XCDE_IL_Couple_Type)) ;
       
	   result_list_dimension*=2 ;
	   }
	 
	 // when there're exactly 2 lists there's a special case, increases also the rightmost position
	 if (number_proximal_lists == 2) 

	   { ++position[rightmost_list_position] ;
	   
	   if (position[rightmost_list_position] >= Proximal_Lists_Length[rightmost_list_position]) 

	     done=1 ;
	   }
	 }

     // if [l, r] is minimal, let r=p, where p is the next element in the leftmost element list
     ++position[leftmost_list_position] ;
     
     if (position[leftmost_list_position] >= Proximal_Lists_Length[leftmost_list_position])
       done=1 ;
   }
 
 if (position != NULL)

   free(position) ;

 return(0) ;
}

// function used to select all minimal windows results that're contained
// in the tag list
int select_proximity_windows_contained_in_tags(XCDE_IL_Couple_Type **selected_IL, 
					       int *selected_IL_num, 
					       XCDE_IL_Couple_Type *tag_IL,
					       int num_el_tag_IL, 
					       XCDE_IL_Couple_Type *windows_results, 
					       int num_el_windows_results) 
{ int pointer_selected_IL=0 ;
 
 int pointer_tag_IL=0 ;
 
 int pointer_windows_results_IL=0 ; 

 // allocates the memory needed to contain the IL 
 *selected_IL = (XCDE_IL_Couple_Type *) malloc((num_el_windows_results)*sizeof(XCDE_IL_Couple_Type)) ;
 
 // loops over the IL 
 while((pointer_tag_IL < num_el_tag_IL) && (pointer_windows_results_IL < num_el_windows_results)) 

   // checks if the window is contained in the tag occurrence
   if (((tag_IL[pointer_tag_IL].XCDE_Position < windows_results[pointer_windows_results_IL].XCDE_Position) ||
	 ((tag_IL[pointer_tag_IL].XCDE_Position == windows_results[pointer_windows_results_IL].XCDE_Position) &&
	  (tag_IL[pointer_tag_IL].XCDE_Offset <= windows_results[pointer_windows_results_IL].XCDE_Offset))) &&
       (tag_IL[pointer_tag_IL+1].XCDE_Position > windows_results[pointer_windows_results_IL+1].XCDE_Position) ||
       ((tag_IL[pointer_tag_IL+1].XCDE_Position == windows_results[pointer_windows_results_IL+1].XCDE_Position) &&
	(tag_IL[pointer_tag_IL+1].XCDE_Offset >= windows_results[pointer_windows_results_IL+1].XCDE_Offset))) { // if enters this block the windows is 
                                                                                                                // contained in the tag occurrence

     // copies the window in the result list
     (*selected_IL)[pointer_selected_IL].XCDE_Position = windows_results[pointer_windows_results_IL].XCDE_Position ;
     
     (*selected_IL)[pointer_selected_IL].XCDE_Offset = windows_results[pointer_windows_results_IL].XCDE_Offset ;
     
     (*selected_IL)[pointer_selected_IL+1].XCDE_Position = windows_results[pointer_windows_results_IL+1].XCDE_Position ;
     
     (*selected_IL)[pointer_selected_IL+1].XCDE_Offset = windows_results[pointer_windows_results_IL+1].XCDE_Offset ;

     // increases the result list pointer
     pointer_selected_IL+=2 ;
     
     // increases the windows list pointer
     pointer_windows_results_IL+=2 ;

     }
    
   else

     // checks if the tag occurrence is after the window 
     if ((tag_IL[pointer_tag_IL].XCDE_Position >= windows_results[pointer_windows_results_IL].XCDE_Position) ||
	 ((tag_IL[pointer_tag_IL].XCDE_Position == windows_results[pointer_windows_results_IL].XCDE_Position) &&
	  (tag_IL[pointer_tag_IL].XCDE_Offset >= windows_results[pointer_windows_results_IL].XCDE_Offset)))
	
       // the tag occurrence is after the window. Increases the windows list pointer 
       pointer_windows_results_IL+=2 ; 

     else
	
       // checks if the tag occurrence is before the window 
       if ((tag_IL[pointer_tag_IL].XCDE_Position <= windows_results[pointer_windows_results_IL].XCDE_Position) ||
	   ((tag_IL[pointer_tag_IL].XCDE_Position == windows_results[pointer_windows_results_IL].XCDE_Position) &&
	    (tag_IL[pointer_tag_IL].XCDE_Offset <= windows_results[pointer_windows_results_IL].XCDE_Offset)))
	 
	 // the tag occurrence is before the window. Increases the tag list pointer
	 pointer_tag_IL+=2 ;
 
 // copies the result list 
 *selected_IL_num=pointer_selected_IL ;

 return(1) ;
}			 
   
     
 
